with integer_io;                        use integer_io;
with Timing_Package;                    use Timing_Package;
with Standard_Complex_Numbers;          use Standard_Complex_Numbers;
with Standard_Random_Numbers;           use Standard_Random_Numbers;
with Standard_Random_Vectors;           use Standard_Random_Vectors;
with Standard_Random_Matrices;
with Standard_Complex_Polynomials;      use Standard_Complex_Polynomials;
with Standard_Complex_Polynomials_io;   use Standard_Complex_Polynomials_io;
with Standard_Complex_Poly_Systems_io;  use Standard_Complex_Poly_Systems_io;
with Standard_Complex_Poly_SysFun;      use Standard_Complex_Poly_SysFun;
with Standard_Complex_Jaco_Matrices;    use Standard_Complex_Jaco_Matrices;
with Standard_Complex_Solutions_io;     use Standard_Complex_Solutions_io;
with Witness_Points;                    use Witness_Points;
with Standard_Plane_Representations;    use Standard_Plane_Representations;
with Standard_Intrinsic_Solutions;      use Standard_Intrinsic_Solutions;
with Extrinsic_Diagonal_Homotopies;     use Extrinsic_Diagonal_Homotopies;
with Diagonal_Homotopy_Continuation;    use Diagonal_Homotopy_Continuation;

package body Drivers_to_Intersect_Varieties is

  function Square_of_the_Original
             ( n,d : natural; ep : Poly_Sys ) return Poly_Sys is

    p : constant Poly_Sys := Remove_Embedding(ep,d);

  begin
    return Complete(n,d,p);
  end Square_of_the_Original;

  procedure Intrinsic_Diagonal_Homotopy
              ( file : in file_type; report : in boolean;
                ep1,ep2 : in Poly_Sys; esols1,esols2 : in Solution_List;
                a,b : in natural ) is

  -- DESCRIPTION :
  --   Diagonal homotopy, assuming a >= b. 

    n : constant natural := Number_of_Unknowns(ep1(ep1'first)) - a;
    k1 : constant natural := n-a;
    k2 : constant natural := n-b;
    s1 : VecVec(1..a) := Slices(ep1,a);
    s2 : VecVec(1..b) := Slices(ep2,b);
    eqs1 : Matrix(1..a,0..n) := Equations_to_Matrix(s1,n);
    eqs2 : Matrix(1..b,0..n) := Equations_to_Matrix(s2,n);
    gen1 : Matrix(1..n,0..k1) := Generators(eqs1);
    gen2 : Matrix(1..n,0..k2) := Generators(eqs2);
    pla1 : Matrix(1..n,0..k1) := Orthogonalize(gen1);
    pla2 : Matrix(1..n,0..k2) := Orthogonalize(gen2);
    isols1 : Solution_List := Project(esols1,pla1);
    isols2 : Solution_List := Project(esols2,pla2);
    p1 : Poly_Sys(1..n-a) := Square_of_the_Original(n,a,ep1);
    p2 : Poly_Sys(1..n-b) := Square_of_the_Original(n,b,ep2);
    sols,esols : Solution_List;
    m : constant natural := k1+k2;
    plane : Matrix(1..n,0..m);

  begin
    Intrinsic_Diagonal_Homotopy
      (file,report,a,b,p1,p2,isols1,isols2,pla1,pla2,sols,plane);
    new_line(file);
    put(file,p1'last+p2'last,1); put(file,"  ");
    put(file,n,1); new_line(file);
    for i in p1'range loop
      put(file," "); put(file,p1(i)); new_line(file);
    end loop;
    for i in p2'range loop
      put(file," "); put(file,p2(i)); new_line(file);
    end loop;
    new_line(file);
    if not Is_Null(sols)
     then put_line(file,"THE SOLUTIONS :");
          esols := Expand(sols,plane);
          put(file,Length_Of(esols),Head_Of(esols).n,esols);
    end if;
  end Intrinsic_Diagonal_Homotopy;

  procedure Intrinsic_Diagonal_Homotopy
              ( file : in file_type; report : in boolean;
                ep1,ep2 : in Poly_Sys; esols1,esols2 : in Solution_List;
                a,b : in natural; f : out Link_to_Poly_Sys;
                p : out Link_to_Matrix; s : out Solution_List ) is

  -- DESCRIPTION :
  --   Diagonal homotopy, assuming a >= b. 

    n : constant natural := Number_of_Unknowns(ep1(ep1'first)) - a;
    k1 : constant natural := n-a;
    k2 : constant natural := n-b;
    s1 : VecVec(1..a) := Slices(ep1,a);
    s2 : VecVec(1..b) := Slices(ep2,b);
    eqs1 : Matrix(1..a,0..n) := Equations_to_Matrix(s1,n);
    eqs2 : Matrix(1..b,0..n) := Equations_to_Matrix(s2,n);
    gen1 : Matrix(1..n,0..k1) := Generators(eqs1);
    gen2 : Matrix(1..n,0..k2) := Generators(eqs2);
    pla1 : Matrix(1..n,0..k1) := Orthogonalize(gen1);
    pla2 : Matrix(1..n,0..k2) := Orthogonalize(gen2);
    isols1 : Solution_List := Project(esols1,pla1);
    isols2 : Solution_List := Project(esols2,pla2);
    p1 : Poly_Sys(1..n-a) := Square_of_the_Original(n,a,ep1);
    p2 : Poly_Sys(1..n-b) := Square_of_the_Original(n,b,ep2);
    sols,esols : Solution_List;
    m : constant natural := k1+k2;
    plane : Matrix(1..n,0..m);

  begin
    Intrinsic_Diagonal_Homotopy
      (file,report,a,b,p1,p2,isols1,isols2,pla1,pla2,sols,plane);
    f := new Poly_Sys(1..p1'last+p2'last);
    for i in p1'range loop
      Copy(p1(i),f(i));
    end loop;
    for i in p2'range loop
      Copy(p2(i),f(p1'last+i));
    end loop;
    p := new Matrix'(plane);
    s := sols;
  end Intrinsic_Diagonal_Homotopy;

  procedure Generic_Diagonal_Homotopy
              ( file : in file_type; nefA,nefB,n,a,b : in natural;
                esols1,esols2 : in Solution_List; p1,p2 : in VecVec ) is

    k1 : constant natural := n-a;
    k2 : constant natural := n-b;
    eqs1 : Matrix(1..a,0..n) := Equations_to_Matrix(p1,n);
    eqs2 : Matrix(1..b,0..n) := Equations_to_Matrix(p2,n);
    gen1 : Matrix(1..n,0..k1) := Generators(eqs1);
    gen2 : Matrix(1..n,0..k2) := Generators(eqs2);
    pla1 : Matrix(1..n,0..k1) := Orthogonalize(gen1);
    pla2 : Matrix(1..n,0..k2) := Orthogonalize(gen2);
    isols1 : Solution_List := Project(esols1,pla1);
    isols2 : Solution_List := Project(esols2,pla2);
    sols : Solution_List;
    m : constant natural := k1+k2;
    plane : Matrix(1..n,0..m);

    procedure Diagonal_Homotopy is new
      Diagonal_Homotopy_Continuation.Generic_Diagonal_Homotopy(fA,jfA,fB,jfB);

  begin
    Diagonal_Homotopy
      (file,nefA,nefB,n,a,b,isols1,isols2,pla1,pla2,sols,plane);
  end Generic_Diagonal_Homotopy;

  function Complete ( r,c : natural; x : Vector; v : Matrix ) return Vector is

  -- DESCRIPTION :
  --   Returns a vector of range 1..r, copying from x(1..r) and adding
  --   components of x with index higher than r, multiplied with v entries.

    res : Vector(1..r);

  begin
    for i in 1..r loop
      res(i) := x(i);
      for j in 1..c loop
        res(i) := res(i) + v(i,j)*x(r+j);
      end loop;
    end loop;
    return res;
  end Complete;

  function Complete ( r,c,n : natural; x,v : Matrix ) return Matrix is

  -- DESCRIPTION :
  --   Returns a matrix with r rows and n columns, copying the first r
  --   rows of x and adding to it entries of x with row indices > r,
  --   multiplied with elements of v.

    res : Matrix(1..r,1..n);

  begin
    for i in 1..r loop
      for k in 1..n loop
        res(i,k) := x(i,k);
        for j in 1..c loop
          res(i,k) := res(i,k) + v(i,j)*x(r+j,k);
        end loop;
      end loop;
    end loop;
    return res;
  end Complete;

  procedure B_Call_Generic_Diagonal_Homotopy
              ( file : in file_type; nef1,n,a,b : in natural;
                ep2 : in Poly_Sys; esols1,esols2 : in Solution_List;
                s1 : in VecVec ) is

    s2 : VecVec(1..b) := Slices(ep2,b);
    p2 : constant Poly_Sys := Remove_Embedding(ep2,b);
    nef2 : constant natural := p2'last;
    f2 : Eval_Poly_Sys(p2'range) := Create(p2);
    jmf2 : Jaco_Mat(p2'range,1..n) := Create(p2);
    ejf2 : Eval_Jaco_Mat(p2'range,1..n) := Create(jmf2);

  begin
    if n-b = nef2 then
      declare
        function fB ( x : Vector ) return Vector is
        begin
         return Eval(f2,x);
        end fB;
        function jfB ( x : Vector ) return Matrix is
        begin
         return Eval(ejf2,x);
        end jfB;
        procedure Diagonal_Homotopy is
          new Generic_Diagonal_Homotopy(fA,jfA,fB,jfB);
      begin
        Diagonal_Homotopy(file,nef1,nef2,n,a,b,esols1,esols2,s1,s2);
      end;
    else
      declare
        r : constant natural := n-b;
        c : constant natural := nef2-r;
        v : Matrix(1..r,1..c) := Standard_Random_Matrices.Random_Matrix(r,c);
        function fB ( x : Vector ) return Vector is
        begin
         return Complete(r,c,Eval(f2,x),v);
        end fB;
        function jfB ( x : Vector ) return Matrix is
        begin
         return Complete(r,c,n,Eval(ejf2,x),v);
        end jfB;
        procedure Diagonal_Homotopy is
          new Generic_Diagonal_Homotopy(fA,jfA,fB,jfB);
      begin
        Diagonal_Homotopy(file,nef1,r,n,a,b,esols1,esols2,s1,s2);
      end;
    end if;
  end B_Call_Generic_Diagonal_Homotopy;

  procedure A_Call_Generic_Diagonal_Homotopy
              ( file : in file_type; ep1,ep2 : in Poly_Sys;
                esols1,esols2 : in Solution_List; a,b : in natural ) is

    n : constant natural := Number_of_Unknowns(ep1(ep1'first)) - a;
    s1 : VecVec(1..a) := Slices(ep1,a);
    p1 : constant Poly_Sys := Remove_Embedding(ep1,a);
    nef1 : constant natural := p1'last;
    f1 : Eval_Poly_Sys(p1'range) := Create(p1);
    jmf1 : Jaco_Mat(p1'range,1..n) := Create(p1);
    ejf1 : Eval_Jaco_Mat(p1'range,1..n) := Create(jmf1);

  begin
    if n-a = nef1 then
      declare
        function fA ( x : Vector ) return Vector is
        begin
          return Eval(f1,x);
        end fA;
        function jfA ( x : Vector ) return Matrix is
        begin
          return Eval(ejf1,x);
        end jfA;
        procedure A_Diagonal_Homotopy is
          new B_Call_Generic_Diagonal_Homotopy(fA,jfA);
      begin
        A_Diagonal_Homotopy(file,nef1,n,a,b,ep2,esols1,esols2,s1);
      end;
    else
      declare
        r : constant natural := n-a;
        c : constant natural := nef1-r;
        v : Matrix(1..r,1..c) := Standard_Random_Matrices.Random_Matrix(r,c);
        function fA ( x : Vector ) return Vector is
        begin
          return Complete(r,c,Eval(f1,x),v);
        end fA;
        function jfA ( x : Vector ) return Matrix is
        begin
          return Complete(r,c,n,Eval(ejf1,x),v);
        end jfA;
        procedure A_Diagonal_Homotopy is
          new B_Call_Generic_Diagonal_Homotopy(fA,jfA);
      begin
        A_Diagonal_Homotopy(file,r,n,a,b,ep2,esols1,esols2,s1);
      end;
    end if;
  end A_Call_Generic_Diagonal_Homotopy;

end Drivers_to_Intersect_Varieties;
