with text_io;                           use text_io;
with Permutations;                      use Permutations;
with Symbol_Table;                      use Symbol_Table;

package Extrinsic_Diagonal_Homotopies_io is

-- DESCRIPTION :
--   This package provides operations to double the symbol table,
---  when diagonal homotopies are used.

  type Symbol_Array is array ( integer range <> ) of Symbol;
  type Link_to_Symbol_Array is access Symbol_Array;

  procedure Write ( sa : in Symbol_Array );
  procedure Write ( file : in file_type; sa : in Symbol_Array ); 

  -- DESCRIPTION :
  --   Writes all symbols in the array to screen or to the file.

  function Add_Suffix ( sb : Symbol; c : character ) return Symbol;
  function Add_Suffix ( sa : Symbol_Array; c : character )
                      return Symbol_Array;

  -- DESCRIPTION :
  --   Adds the character c to the symbol sb or to the symbols in sa.

  function Remove_Suffix ( sb : Symbol ) return Symbol;

  -- DESCRIPTION :
  --   Removes the last character from the symbol.

  function Get_Symbols return Symbol_Array;
  function Get_Link_to_Symbols return Link_to_Symbol_Array;

  -- DESCRIPTION :
  --   Returns the array with all symbols from the symbol table.

  function Suffixed_Symbols ( c : character ) return Symbol_Array;

  -- DESCRIPTION :
  --   Returns an array with all symbols from the symbol table,
  --   appended with the suffix c.

  function Is_Embed_Symbol ( sb : Symbol ) return boolean;

  -- DESCRIPTION :
  --   Returns true if the symbol is a symbol used in the embedding,
  --   i.e.: whether the first two characters are z.

  function Remove_Embed_Symbols ( sa : Symbol_Array ) return Symbol_Array;

  -- DESCRIPTION :
  --   Returns the array with embedding symbols removed from sa.

  function Equals_mod_Suffix ( sb1,sb2 : Symbol ) return boolean;

  -- DESCRIPTION :
  --   Returns true if the symbols are the same, disregarding the
  --   last character.

  function Look_for_Position 
              ( sa : Symbol_Array; sb : Symbol ) return natural;

  -- DESCRIPTION :
  --   Returns 0 if the symbol sb does not occur in the array sa,
  --   otherwise the position of the symbol in the array is returned.

  function Search_Position 
              ( sa : Symbol_Array; sb : Symbol ) return natural;

  -- DESCRIPTION :
  --   Returns 0 if the symbol sb does not occur in the array sa,
  --   otherwise the position of the symbol in the array is returned,
  --   suffixes are ignored.

  function Match_Symbols
              ( s1,s2 : in Symbol_Array ) return Permutation;

  -- DESCRIPTION :
  --   Matches the symbols in second array with those in the first.
  --   Creates a permutation, for every symbol in s2, the i-th position
  --   in the permutation equals the position of the same symbol in s1.

  function Matching_Permutation
              ( s1,s2 : in Symbol_Array ) return Permutation;

  -- DESCRIPTION :
  --   Matches the symbols in second array with those in the first.
  --   Creates a permutation, for every symbol in s2, the i-th position
  --   in the permutation equals the position of the same symbol in s1.
  --   The suffixes (i.e.: last character of every symbol) are ignored.

  -- REQUIRED : s1'range = s2'range.

  procedure Permute ( prm : in Permutation; s : in out Symbol_Array );

  -- DESCRIPTION :
  --   Permutes the symbols in the array according the permutation.

  procedure Write_Symbol_Table;

  -- DESCRIPTION :
  --   Writes the current contents of the symbol table to screen.

  procedure Assign_Symbol_Table ( sa : in Symbol_Array );

  -- DESCRIPTION :
  --   Assigns the symbol table with the symbols in the symbol array.

  procedure Assign_Symbol_Table ( s1,s2 : in Symbol_Array );

  -- DESCRIPTION :
  --   Assigns the symbol table with the symbols in the symbol arrays.

  procedure Clear ( ls : in out Link_to_Symbol_Array );

  -- DESCRIPTION :
  --   Releases the memory occupied by the symbol array.

end Extrinsic_Diagonal_Homotopies_io;
