with text_io;                           use text_io;
with Standard_Floating_Numbers;         use Standard_Floating_Numbers;
with Standard_Complex_Vectors;          use Standard_Complex_Vectors;
with Standard_Complex_Matrices;         use Standard_Complex_Matrices;

package Standard_Cascading_Planes is

-- DESCRIPTION :
--   Provides utilities to build a sequence of linear spaces,
--   for an intrinsic diagonal homotopy method to intersect
--   positive dimensional varieties.

  function Project ( x : Matrix; h : natural ) return Matrix;

  -- DESCRIPTION :
  --   Copies the first h rows of x and sets the other rows to zero.
  --   The matrix on return has the same dimensions as x.

  function Project ( x : Vector; h : natural ) return Vector;

  -- DESCRIPTION :
  --   Copies the first h rows of x and sets the other rows to zero.
  --   The vector on return has the same dimensions as x.

  function Double_to_Diagonal ( A : Matrix ) return Matrix;

  -- DESCRIPTION :
  --   Doubles the number of columns of A, according to the
  --   equations on the diagonal.

  function Compute_Offset ( C : Matrix; d : Vector ) return Vector;

  -- DESCRIPTION :
  --   Returns the solution w to the system C*w + d = 0.

  procedure Shift_Offset ( p : in out Matrix; b : in Vector );

  -- DESCRIPTION :
  --   Updates the generators of the plane p with a new offset b.

  procedure Evaluate ( file : in file_type;
                       equ : in Matrix; v : in Vector;
                       res : out double_float );
  procedure Evaluate ( equ : in Matrix; v : in Vector;
                       res : out double_float );

  -- DESCRIPTION :
  --   Evaluates the vector v at the equations in equ.
  --   The 1-norm of the evaluation is in res.
  --   Intermediate output is written to file.

  procedure Evaluate ( file : in file_type;
                       p : in Matrix; g : in Matrix;
                       res : out double_float );
  procedure Evaluate ( p : in Matrix; g : in Matrix;
                       res : out double_float );

  -- DESCRIPTION :
  --   Verifies whether the generators in g satisfy the equations in p.
  --   The 1-norm of the evaluation is in res.
  --   Intermediate output is written to file.

  procedure Intersect ( e1,e2 : in Matrix; p1,p2 : in out Matrix );
  procedure Intersect ( file : in file_type;
                        e1,e2 : in Matrix; p1,p2 : in out Matrix );

  -- DESCRIPTION :
  --   Computes the intersection of two planes.

  -- ON ENTRY :
  --   file      for writing diagnostics to, if file is provided;
  --   e1,e2     equations for the two planes;
  --   p1,p2     orthogonal representations of the two planes,
  --             with common offset vector.

  -- ON RETURN :
  --   p1,p2     common basis, except for one complementary vector,
  --             not in the intersection of both planes.

  function Target_Space ( n,n2,apb,b : natural; dA,BB,CC : Matrix;
                          d : Vector ) return Matrix;

  -- DESCRIPTION :
  --   Generates the equations for the target linear space.

  -- ON ENTRY :
  --   n         dimension of the ambient space before embedding;
  --   n2        equals 2*n;
  --   apb       is a + b, sum of dimension of the two components;
  --   b         smallest dimension of the two components;
  --   dA        matrix of a+b rows and n2 columns, randomized diagonal;
  --   BB        matrix of a+b rows and n2 columns, is a multiplier;
  --   CC,d      defines the system CC*w + d = 0.

  -- ON RETURN :
  --   matrix of a+b rows and columns of range 0..n2.

  function Start_Space ( g1,g2 : Matrix ) return Matrix;

  -- DESCRIPTION :
  --   Returns an intrinsic representation of the plane at the start,
  --   from the generators of the planes cutting the witness sets of
  --   the two components.

end Standard_Cascading_Planes;
