with Standard_Floating_Numbers;        use Standard_Floating_Numbers;
with Standard_Complex_Numbers;         use Standard_Complex_Numbers;
with Standard_Complex_Solutions;       use Standard_Complex_Solutions;

package Standard_Solution_Filters is

-- DESCRIPTION :
--   This package offers tools to create new solution lists by
--   filtering lists subject to certain criteria.

-- CRITERIA :

  function Vanishing ( sol : Solution; tol : double_float ) return boolean;

  -- DESCRIPTION :
  --   Returns true if either the error on the last correction term
  --   or the residual are smaller than the given tolerance.

  function Zero_Component ( sol : Solution; k : natural;
                            tol : double_float ) return boolean;

  -- DESCRIPTION :
  --   The k-th component of a solution is considered as zero if its
  --   absolute value is less than the given tolerance.

  function Regular ( sol : Solution; tol : double_float ) return boolean;

  -- DESCRIPTION :
  --   A solution is regular if the estimate for the inverse condition
  --   number is larger than the given tolerance.

  function On_Target ( sol : Solution; target : Complex_Number;
                       tol : double_float ) return boolean;

  -- DESCRIPTION :
  --   Returns true if | sol.t - target | < tol.

-- FILTERS :

  function Vanishing_Filter ( sols : Solution_List; tol : double_float )
                            return Solution_List;

  -- DESCRIPTION :
  --   Returns the list of solutions that are considered as vanishing
  --   with respect to the given tolerance on errors and residuals.

  function Spurious_Filter ( sols : Solution_List; tol : double_float )
                           return Solution_List;

  -- DESCRIPTION :
  --   Returns the list of solutions that are considered as spurious
  --   that is nonvanishing for the given tolerance.

  function Zero_Component_Filter
               ( sols : Solution_List; k : natural; tol : double_float )
               return Solution_List;

  -- DESCRIPTION :
  --   Returns the list of solutions that are considered as having their
  --   k-th component zero with respect to the given tolerance.

  function Free_Component_Filter
               ( sols : Solution_List; k : natural; tol : double_float )
               return Solution_List;

  -- DESCRIPTION :
  --   Returns the list of solutions that are considered as having their
  --   k-th component free (nonzero) with respect to the given tolerance.

  function Regular_Filter ( sols : Solution_List; tol : double_float )
                          return Solution_List;

  -- DESCRIPTION :
  --   Returns the list of solutions that are considered as regular with
  --   respect to the given tolerance for the inverse condition number.

  function Singular_Filter ( sols : Solution_List; tol : double_float )
                           return Solution_List;

  -- DESCRIPTION :
  --   Returns the list of solutions that are considered as singular with
  --   respect to the given tolerance for the inverse condition number.

  function On_Target_Filter ( sols : Solution_List; target : Complex_Number;
                              tol : double_float ) return Solution_List;

  -- DESCRIPTION :
  --   Returns the list of solutions that have reached the given target.

  function Off_Target_Filter ( sols : Solution_List; target : Complex_Number;
                               tol : double_float ) return Solution_List;

  -- DESCRIPTION :
  --   Returns the list of solutions that have not reached the given target.

end Standard_Solution_Filters;
