with integer_io;                         use integer_io;
with Standard_Floating_Numbers;          use Standard_Floating_Numbers;
with Standard_Floating_Numbers_io;       use Standard_Floating_Numbers_io;
with Standard_Complex_Numbers;           use Standard_Complex_Numbers;
with Standard_Complex_Numbers_io;        use Standard_Complex_Numbers_io;
with Standard_Complex_Numbers_Polar;     use Standard_Complex_Numbers_Polar;
with Standard_Random_Numbers;            use Standard_Random_Numbers;
with Standard_Natural_Vectors;
with Standard_Random_Vectors;            use Standard_Random_Vectors;
with Standard_Complex_Norms_Equals;      use Standard_Complex_Norms_Equals;
with Standard_Complex_Poly_SysFun;       use Standard_Complex_Poly_SysFun;
with Lists_of_Complex_Vectors;           use Lists_of_Complex_Vectors;
with Hypersurface_Samplers;              use Hypersurface_Samplers;
with Monodromy_Partitions;               use Monodromy_Partitions;

package body Monodromy_Polynomial_Breakup is

  tol : constant double_float := 1.0E-8;
  threshold : constant natural := 10;

-- AUXILIARY OPERATIONS :

  function Good_Map ( mp : Standard_Natural_Vectors.Vector )
                    return boolean is

  -- DESCRIPTION :
  --   A good map contains no zero elements.

  begin
    for i in mp'range loop
      if mp(i) = 0
       then return false;
      end if;
    end loop;
    return true;
  end Good_Map;

  function Eval_Sys ( n : natural; p : Poly; ep : Eval_Poly )
                    return Eval_Poly_Sys is

  -- DESCRIPTION :
  --   The array on return has range 0..n, where n is the number
  --   of variables in the polynomial p.  The 0-th entry is ep,
  --   while the i-th entry is the i-th derivative of p.

    res : Eval_Poly_Sys(0..n);
    dp : Poly;

  begin
    res(0) := ep;
    for i in 1..n loop
      dp := Diff(p,i);
      res(i) := Create(dp);
      Clear(dp);
    end loop;
    return res;
  end Eval_Sys;

  procedure Append_Line 
              ( bl,bl_last,vl,vl_last,tl,tl_last : in out List;
                b,v,t : in Vector ) is

  -- DESCRIPTION :
  --   A line consists of a base point b, direction v and roots t.

  -- ON ENTRY :
  --   (bl,bl_last)    first and last element of list of base points;
  --   (vl,vl_last)    first and last element of list of directions;
  --   (tl,tl_last)    first and last element of list of roots;
  --   (b,v,t)         new line.

  -- ON RETURN :
  --   (bl,bl_last)    updated list of base points;
  --   (vl,vl_last)    updated list of directions;
  --   (tl,tl_last)    updated list of roots.

  begin
    Append(bl,bl_last,b);
    Append(vl,vl_last,v);
    Append(tl,tl_last,t);
  end Append_Line;

  procedure First_Loop
              ( fp : in Eval_Poly_Sys;
                bl,bl_last,vl,vl_last,tl,tl_last : in out List;
                b0,v0,t0 : in Standard_Complex_Vectors.Vector;
                deco : in out Link_to_VecVec; nb : in out natural;
                fail : out boolean ) is

  -- DESCRIPTION :
  --   Executes the first monodromy loop, from the line (b0,v0) with
  --   the roots in t0.  Initializes the lists (see the Append_Line
  --   above for their meaning) with three lines and their roots.
  --   On return is the map connecting points on the same factor.
  --   If fail, then the map was found to contain zeros.

    eps : constant double_float := 1.0E-13;
    max : constant natural := 4;
    b1 : Vector(b0'range) := Random_Vector(b0'first,b0'last);
    v1 : Vector(v0'range) := Random_Vector(v0'first,v0'last);
    b2 : Vector(b0'range) := Random_Vector(b0'first,b0'last);
    v2 : Vector(v0'range) := Random_Vector(v0'first,v0'last);
    t : Vector(t0'range) := t0;
    t1,t2 : Vector(t0'range);
    ft,dt : Vector(t'range);
    mp : Standard_Natural_Vectors.Vector(t'range);

  begin
   -- b1(1) := Create(0.0);
   -- b2(1) := Create(0.0);
   -- for i in v0'first+1..v0'last loop
   --   v1(i) := Create(0.0);
   --   v2(i) := Create(0.0);
   -- end loop;
    Silent_Hypersurface_Sampler(fp,b0,v0,b1,v1,t);
    Silent_Refiner(fp,b1,v1,t,ft,dt,eps,max);
    t1 := t;
    Silent_Hypersurface_Sampler(fp,b1,v1,b2,v2,t);
    Silent_Refiner(fp,b2,v2,t,ft,dt,eps,max);
    t2 := t;
    Silent_Hypersurface_Sampler(fp,b2,v2,b0,v0,t);
    Silent_Refiner(fp,b0,v0,t,ft,dt,eps,max);
    mp := Map(t0,t,tol);
    fail := not Good_Map(mp);
    if not fail
     then Append_Line(bl,bl_last,vl,vl_last,tl,tl_last,b0,v0,t0);
          Append_Line(bl,bl_last,vl,vl_last,tl,tl_last,b1,v1,t1);
          Append_Line(bl,bl_last,vl,vl_last,tl,tl_last,b2,v2,t2);
          nb := deco'last;
          Add_Map(deco,nb,mp);
    end if;
  end First_Loop;

  procedure First_Loop
              ( file : in file_type;
                fp : in Eval_Poly_Sys; output : in boolean;
                bl,bl_last,vl,vl_last,tl,tl_last : in out List;
                b0,v0,t0 : in Standard_Complex_Vectors.Vector;
                deco : in out Link_to_VecVec; nb : in out natural;
                fail : out boolean ) is

  -- DESCRIPTION :
  --   Executes the first monodromy loop, from the line (b0,v0) with
  --   the roots in t0.  Initializes the lists (see the Append_Line
  --   above for their meaning) with three lines and their roots.
  --   On return is the updated connectivity information.

    eps : constant double_float := 1.0E-13;
    max : constant natural := 4;
    b1 : Vector(b0'range) := Random_Vector(b0'first,b0'last);
    v1 : Vector(v0'range) := Random_Vector(v0'first,v0'last);
    b2 : Vector(b0'range) := Random_Vector(b0'first,b0'last);
    v2 : Vector(v0'range) := Random_Vector(v0'first,v0'last);
    t : Vector(t0'range) := t0;
    t1,t2 : Vector(t0'range);
    ft,dt : Vector(t'range);
    mp : Standard_Natural_Vectors.Vector(t'range);

  begin
   -- b1(1) := Create(0.0);
   -- b2(1) := Create(0.0);
   -- for i in v0'first+1..v0'last loop
   --   v1(i) := Create(0.0);
   --   v2(i) := Create(0.0);
   -- end loop;
    Reporting_Hypersurface_Sampler(file,fp,b0,v0,b1,v1,output,t);
   -- Reporting_Refiner(file,fp,b1,v1,t,ft,dt,eps,max);
    Silent_Refiner(fp,b1,v1,t,ft,dt,eps,max);
    t1 := t;
    Reporting_Hypersurface_Sampler(file,fp,b1,v1,b2,v2,output,t);
   -- Reporting_Refiner(file,fp,b2,v2,t,ft,dt,eps,max);
    Silent_Refiner(fp,b2,v2,t,ft,dt,eps,max);
    t2 := t;
    Reporting_Hypersurface_Sampler(file,fp,b2,v2,b0,v0,output,t);
   -- Reporting_Refiner(file,fp,b0,v0,t,ft,dt,eps,max);
    Silent_Refiner(fp,b0,v0,t,ft,dt,eps,max);
    mp := Map(t0,t,tol);
    Write_Map(file,mp);
    fail := not Good_Map(mp);
    if fail
     then put_line(file,"The map is not good for processing.");
     else Append_Line(bl,bl_last,vl,vl_last,tl,tl_last,b0,v0,t0);
          Append_Line(bl,bl_last,vl,vl_last,tl,tl_last,b1,v1,t1);
          Append_Line(bl,bl_last,vl,vl_last,tl,tl_last,b2,v2,t2);
          nb := deco'last;
          Add_Map(deco,nb,mp);
    end if;
  end First_Loop;

  procedure New_Loop
              ( fp : in Eval_Poly_Sys; d : in natural;
                bl,bl_last,vl,vl_last,tl,tl_last : in out List;
                b0,v0,t0 : in Standard_Complex_Vectors.Vector;
                deco : in out Link_to_VecVec; nb,scnt : in out natural;
                fail : out boolean ) is

  -- DESCRIPTION :
  --   Attempts to find a new monodromy group action.

    eps : constant double_float := 1.0E-13;
    max : constant natural := 4;
    b : Vector(b0'range) := Random_Vector(b0'first,b0'last);
    v : Vector(v0'range) := Random_Vector(v0'first,v0'last);
    t : Vector(t0'range) := t0;
    wb,wv : Vector(b'range);
    ft,dt,wt : Vector(t'range);
    mp : Standard_Natural_Vectors.Vector(t'range);
    tmpbl : List := Tail_Of(bl);
    tmpvl : List := Tail_Of(vl);
    tmptl : List := Tail_Of(tl);
    prvnb : natural := nb;
    gm : boolean;

  begin
   -- b(1) := Create(0.0);
   -- for i in v0'first+1..v0'last loop
   --   v(i) := Create(0.0);
   -- end loop;
    fail := true;
    Silent_Hypersurface_Sampler(fp,b0,v0,b,v,t);
    Silent_Refiner(fp,b,v,t,ft,dt,eps,max);
    while not Is_Null(tmpbl) loop
      wb := Head_Of(tmpbl).all;
      wv := Head_Of(tmpvl).all;
      wt := t;
      Silent_Hypersurface_Sampler(fp,b,v,wb,wv,wt);
      Silent_Refiner(fp,wb,wv,wt,ft,dt,eps,max);
      mp := Map(Head_Of(tmptl).all,wt,tol);
      gm := Good_Map(mp);
      if gm
       then Add_Map(deco,nb,mp);
            fail := false;
       else return;
      end if;
      exit when (nb = 1);
      if gm
       then if prvnb = nb
             then scnt := scnt + 1;
             else prvnb := nb;
                  scnt := 0;
            end if;
      end if;
      exit when (scnt >= threshold);
      tmpbl := Tail_Of(tmpbl);
      tmpvl := Tail_Of(tmpvl);
      tmptl := Tail_Of(tmptl);
    end loop;
    if not fail
     then Append_Line(bl,bl_last,vl,vl_last,tl,tl_last,b,v,t);
    end if;
  end New_Loop;

  procedure New_Loop
              ( file : in file_type; d : in natural;
                fp : in Eval_Poly_Sys; output : in boolean;
                bl,bl_last,vl,vl_last,tl,tl_last : in out List;
                b0,v0,t0 : in Standard_Complex_Vectors.Vector;
                deco : in out Link_to_VecVec; nb,scnt : in out natural;
                fail : out boolean ) is

  -- DESCRIPTION :
  --   Attempts to find a new monodromy group action.

    eps : constant double_float := 1.0E-14;
    max : constant natural := 4;
    b : Vector(b0'range) := Random_Vector(b0'first,b0'last);
    v : Vector(v0'range) := Random_Vector(v0'first,v0'last);
    t : Vector(t0'range) := t0;
    wb,wv : Vector(b'range);
    ft,dt,wt : Vector(t'range);
    mp : Standard_Natural_Vectors.Vector(t'range);
    tmpbl : List := Tail_Of(bl);
    tmpvl : List := Tail_Of(vl);
    tmptl : List := Tail_Of(tl);
    prvnb : natural := nb;
    gm : boolean;

  begin
   -- b(1) := Create(0.0);
   -- for i in v0'first+1..v0'last loop
   --   v(i) := Create(0.0);
   -- end loop;
    Reporting_Hypersurface_Sampler(file,fp,b0,v0,b,v,output,t);
   -- Reporting_Refiner(file,fp,b,v,t,ft,dt,eps,max);
    Silent_Refiner(fp,b,v,t,ft,dt,eps,max);
    fail := true;
    while not Is_Null(tmpbl) loop
      wb := Head_Of(tmpbl).all;
      wv := Head_Of(tmpvl).all;
      wt := t;
      Reporting_Hypersurface_Sampler(file,fp,b,v,wb,wv,output,wt);
     -- Reporting_Refiner(file,fp,wb,wv,wt,ft,dt,eps,max);
      Silent_Refiner(fp,wb,wv,wt,ft,dt,eps,max);
      mp := Map(Head_Of(tmptl).all,wt,tol);
      Write_Map(file,mp);
      gm := Good_Map(mp);
      if gm
       then Add_Map(deco,nb,mp);
            fail := false;
       else return;
      end if;
      exit when (nb = 1);
      if gm
       then if prvnb = nb
             then scnt := scnt + 1;
             else prvnb := nb;
                  scnt := 0;
            end if;
      end if;
      exit when (scnt > threshold);
      tmpbl := Tail_Of(tmpbl);
      tmpvl := Tail_Of(tmpvl);
      tmptl := Tail_Of(tmptl);
    end loop;
    if not fail
     then Append_Line(bl,bl_last,vl,vl_last,tl,tl_last,b,v,t);
    end if;
  end New_Loop;

-- TARGET PROCEDURES :

  procedure Monodromy_Breakup
                ( file : in file_type; 
                  p : in Poly; ep : in Eval_Poly; b,v,gp : in Vector;
                  output : in boolean; deco : in out Link_to_VecVec ) is

    cnt,nb : natural := 0;
    d : constant natural := gp'length; 
    fp : Eval_Poly_Sys(0..b'last) := Eval_Sys(b'last,p,ep);
    bl,bl_last,vl,vl_last,tl,tl_last : List;
    fail : boolean := true;

  begin
    while fail loop
      First_Loop(file,fp,output,bl,bl_last,vl,vl_last,tl,tl_last,b,v,
                 gp,deco,nb,fail);
    end loop;
    if deco /= null
     then nb := Number_of_Factors(deco.all);
     else nb := 0;
    end if;
    if nb > 1
     then loop
            New_Loop(file,d,fp,output,bl,bl_last,vl,vl_last,tl,tl_last,
                     b,v,gp,deco,nb,cnt,fail);
            nb := Number_of_Factors(deco.all);
            exit when (nb = 1) or (cnt >= threshold);
          end loop;
    end if;
  end Monodromy_Breakup;

  procedure Monodromy_Breakup
                ( p : in Poly; ep : in Eval_Poly; b,v,gp : in Vector;
                  deco : in out Link_to_VecVec ) is

    cnt,nb : natural := 0;
    d : constant natural := gp'length;
    fp : Eval_Poly_Sys(0..b'last) := Eval_Sys(b'last,p,ep);
    bl,bl_last,vl,vl_last,tl,tl_last : List;
    fail : boolean := true;

  begin
    while fail loop
      First_Loop(fp,bl,bl_last,vl,vl_last,tl,tl_last,b,v,gp,deco,nb,fail);
    end loop;
    if deco /= null
     then nb := Number_of_Factors(deco.all);
     else nb := 0;
    end if;
    if nb > 1
     then loop
            New_Loop(fp,d,bl,bl_last,vl,vl_last,tl,tl_last,b,v,
                     gp,deco,nb,cnt,fail);
            nb := Number_of_Factors(deco.all);
            exit when (nb = 1) or (cnt >= threshold);
          end loop;
    end if;
  end Monodromy_Breakup;

end Monodromy_Polynomial_Breakup;
