with integer_io;                        use integer_io;
with Standard_Floating_Numbers;         use Standard_Floating_Numbers;
with Standard_Floating_Numbers_io;      use Standard_Floating_Numbers_io;
with Standard_Complex_Numbers_Polar;    use Standard_Complex_Numbers_Polar;
with Standard_Complex_Numbers_io;       use Standard_Complex_Numbers_io;
with Standard_Complex_Vectors_io;       use Standard_Complex_Vectors_io;
with Standard_Complex_Singular_Values;  use Standard_Complex_Singular_Values;
with Process_io;                        use Process_io;
with Standard_Intrinsic_Newton;         use Standard_Intrinsic_Newton;

package body Standard_Intrinsic_Trackers is

-- AUXILIARIES :

  function At_End ( s : Solu_Info; t1 : Complex_Number;
                    max : natural ) return boolean is

  -- DESCRIPTION :
  --   We are at the end of the path if s.sol.t equals t1,
  --   or if s.niter equals or exceeds max.

  begin
    return (Equal(s.sol.t,t1) or (s.niter > max));
  end At_End;

  procedure Step_Control
              ( fail : in boolean; pp : in Pred_Pars;
                step : in out double_float; nbsuccess : in out natural;
                t,pt0,pt1,pt2,pt3 : in out Complex_Number;
                x,px0,px1,px2,px3 : in out Vector ) is

  -- DESCRIPTION :
  --   Applies step control and management of data for step back.

  -- ON ENTRY :
  --   fail     if true, corrector step failed, otherwise if false;
  --   pp       settings for predictor values;
  --   step     current step size;
  --   nbsucces is the number of consecutive corrector successes;
  --   t        current value of continuation parameter;
  --   pt0      previous value of t where corrector had success;
  --   pt1      value previous to pt0 where corrector had success;
  --   pt2      value previous to pt1 where corrector had success;
  --   pt3      value previous to pt2 where corrector had success;
  --   x        current solution vector corresponding to t;
  --   px0      current solution vector corresponding to pt0;
  --   px1      current solution vector corresponding to pt1;
  --   px2      current solution vector corresponding to pt2;
  --   px3      current solution vector corresponding to pt3.

  -- ON RETURN :
  --   step     updated step size;
  --   nbsucces is updated accordingly to fail;
  --   t        equals pt0 when fail, otherwise is unchanged;
  --   pt0      equals t when not fail, otherwise is unchanged;
  --   pt1      equals pt0 when not fail, otherwise is unchanged;
  --   pt2      equals pt1 when not fail, otherwise is unchanged;
  --   pt3      equals pt2 when not fail, otherwise is unchanged;
  --   px0      equals x when not fail, otherwise is unchanged;
  --   px1      equals px0 when not fail, otherwise is unchanged;
  --   px2      equals px1 when not fail, otherwise is unchanged;
  --   px3      equals px2 when not fail, otherwise is unchanged.

  begin
    if fail
     then nbsuccess := 0;
          step := step*pp.redfac;
     else nbsuccess := nbsuccess + 1;
          if nbsuccess > pp.success_steps
           then step := step*pp.expfac;
                if step > pp.maxstep
                 then step := pp.maxstep;
                end if;
          end if;
    end if;
    if step <= pp.minstep
     then return;
    end if;
    if fail
     then t := pt0; x := px0;
          --t := pt1; x := px1;      -- 01/13 why this ?
          --pt0 := pt1; px0 := px1;
     else pt3 := pt2; px3 := px2;
          pt2 := pt1; px2 := px1;
          pt1 := pt0; px1 := px0;
          pt0 := t;   px0 := x;
    end if;
  end Step_Control;

  procedure Step_Control
              ( file : in file_type; fail : in boolean; pp : in Pred_Pars;
                step : in out double_float; nbsuccess : in out natural;
                t,pt0,pt1,pt2,pt3 : in out Complex_Number;
                x,px0,px1,px2,px3 : in out Vector ) is

  -- DESCRIPTION :
  --   Applies step control and management of data for step back.

  -- ON ENTRY :
  --   fail     if true, corrector step failed, otherwise if false;
  --   pp       settings for predictor values;
  --   step     current step size;
  --   nbsucces is the number of consecutive corrector successes;
  --   t        current value of continuation parameter;
  --   pt0      previous value of t where corrector had success;
  --   pt1      value previous to pt0 where corrector had success;
  --   pt2      value previous to pt1 where corrector had success;
  --   pt3      value previous to pt2 where corrector had success;
  --   x        current solution vector corresponding to t;
  --   px0      current solution vector corresponding to pt0;
  --   px1      current solution vector corresponding to pt1;
  --   px2      current solution vector corresponding to pt2;
  --   px3      current solution vector corresponding to pt3.

  -- ON RETURN :
  --   step     updated step size;
  --   nbsucces is updated accordingly to fail;
  --   t        equals pt0 when fail, otherwise is unchanged;
  --   pt0      equals t when not fail, otherwise is unchanged;
  --   pt1      equals pt0 when not fail, otherwise is unchanged;
  --   pt2      equals pt1 when not fail, otherwise is unchanged;
  --   pt3      equals pt2 when not fail, otherwise is unchanged;
  --   px0      equals x when not fail, otherwise is unchanged;
  --   px1      equals px0 when not fail, otherwise is unchanged;
  --   px2      equals px1 when not fail, otherwise is unchanged;
  --   px3      equals px2 when not fail, otherwise is unchanged.

  begin
    if fail
     then nbsuccess := 0;
          step := step*pp.redfac;
     else nbsuccess := nbsuccess + 1;
          if nbsuccess > pp.success_steps
           then step := step*pp.expfac;
                if step > pp.maxstep
                 then step := pp.maxstep;
                end if;
          end if;
    end if;
    put(file,"#consecutive successes : ");
    put(file,nbsuccess,1); new_line(file);
    if step <= pp.minstep
     then return;
    end if;
    if fail
     then t := pt0; x := px0;
         -- put_line(file,"Back up solution to vector :");
         -- put_line(file,x);
          --t := pt1; x := px1;      -- 01/13 why this ?
          --pt0 := pt1; px0 := px1;
     else pt3 := pt2; px3 := px2;
          pt2 := pt1; px2 := px1;
          pt1 := pt0; px1 := px0;
          pt0 := t;   px0 := x;
    end if;
   -- put_line(file,"Current vector of t-values : ");
   -- put(file,"t3 = "); put(file,pt3); new_line(file);
   -- put(file,"t2 = "); put(file,pt2); new_line(file);
   -- put(file,"t1 = "); put(file,pt1); new_line(file);
   -- put(file,"t0 = "); put(file,pt0); new_line(file);
  end Step_Control;

  function Quadratic_Predictor
               ( t,t0,t1,t2 : Complex_Number; x0,x1,x2 : Vector )
               return Vector is

  -- DESCRIPTION :
  --   Uses extrapolation through (t0,x0), (t1,x1), and (t2,x2) to
  --   predict the solution vector at t.

    v : Vector(x0'range);
    f01,f02,f12,dt01,dt02,dt12,pt0,pt01 : Complex_Number;

  begin
    dt01 := t1 - t0;
    dt02 := t2 - t0;
    dt12 := t2 - t1;
    pt0 := t - t0;
    pt01 := pt0*(t-t1);
    for i in v'range loop
      f01 := (x1(i) - x0(i))/dt01;
      f02 := (x2(i) - x0(i))/dt02;
      f12 := (f02 - f01)/dt12;
      v(i) := x0(i) + f01*pt0 + f12*pt01;
    end loop;
    return v;
  end Quadratic_Predictor;

  function Cubic_Predictor
               ( t,t0,t1,t2,t3 : Complex_Number; x0,x1,x2,x3 : Vector )
               return Vector is

  -- DESCRIPTION :
  --   Uses extrapolation through (t0,x0), (t1,x1), (t2,x2), and (t3,x3)
  --   to predict the solution vector at t.

    v : Vector(x0'range);
    f1,f2,f3,pt0,pt1,pt2,dt01,dt02,dt03,dt12,dt13,dt23 : Complex_Number;

  begin
    pt0 := t-t0;
    pt1 := pt0*(t-t1);
    pt2 := pt1*(t-t2);
    dt01 := t1-t0; dt02 := t2-t0; dt03 := t3-t0;
    dt12 := t2-t1; dt13 := t3-t1; dt23 := t3-t2;
    for i in v'range loop
      f1 := (x1(i) - x0(i))/dt01;
      f2 := (x2(i) - x0(i))/dt02;
      f3 := (x3(i) - x0(i))/dt03;
      f2 := (f2 - f1)/dt12;
      f3 := (f3 - f1)/dt13;
      f3 := (f3 - f2)/dt23;
      v(i) := x0(i) + f1*pt0 + f2*pt1 + f3*pt2;
    end loop;
    return v;
  end Cubic_Predictor;

  procedure Predictor
               ( t : in out Complex_Number; v : in out Vector;
                 step : in out double_float;
                 pt0,pt1,pt2,pt3,t1 : in Complex_Number;
                 px1,px2,px3 : in Vector; fail : in boolean ) is

  -- DESCRIPTION :
  --   Predicts new value for continuation parameter t and,
  --   if not fail, also predicts the corresponding solution vector v.  

  -- ON ENTRY :
  --   t         current value of the continuation parameter t;
  --   v         current value of the solution vector v;
  --   step      current step size;
  --   pt0       backup value for the continuation parameter t;
  --   pt1       value for t previous to backup for t;
  --   pt2       value for t previous to pt1;
  --   pt3       value for t previous to pt2;
  --   t1        target value for continuation parameter (t1 = 1);
  --   px1       previous value for solution vector v;
  --   px2       value for solution vector v corresponding to pt2;
  --   px3       value for solution vector v corresponding to pt3;
  --   fail      true if last correction was not successful.

  -- ON RETURN :
  --   t         predicted value for the continuation parameter t;
  --   v         predicted value for the solution vector v;
  --   step      is actual step size taken.

    d0,d1,d2 : double_float;

  begin
    t := t + Create(step);
    if REAL_PART(t) > 1.0
     then step := REAL_PART(t) - 1.0;
          t := t1;
    end if;
    if not fail then
      d0 := AbsVal(pt0 - pt1);
      if 1.0 + d0 /= 1.0 then
        d1 := AbsVal(pt1 - pt2);
        if 1.0 + d1 = 1.0
         then d0 := step/d0;
              v := v + Create(d0)*(v - px1);
         else d2 := AbsVal(pt2 - pt3);
              if 1.0 + d2 = 1.0
               then v := Quadratic_Predictor(t,pt0,pt1,pt2,v,px1,px2);
               else v := Cubic_Predictor(t,pt0,pt1,pt2,pt3,v,px1,px2,px3);
              end if;
        end if;
      end if;
    end if;
  end Predictor;

  procedure Predictor
               ( file : in file_type;
                 t : in out Complex_Number; v : in out Vector;
                 step : in out double_float;
                 pt0,pt1,pt2,pt3,t1 : in Complex_Number;
                 px1,px2,px3 : in Vector; fail : in boolean ) is

  -- DESCRIPTION :
  --   Predicts new value for continuation parameter t and,
  --   if not fail, also predicts the corresponding solution vector v.  

  -- ON ENTRY :
  --   file      for intermediate diagnostics;
  --   t         current value of the continuation parameter t;
  --   v         current value of the solution vector v;
  --   step      current step size;
  --   pt0       backup value for the continuation parameter t;
  --   pt1       value for t previous to backup for t;
  --   pt2       value for t previous to pt1;
  --   pt3       value for t previous to pt2;
  --   t1        target value for continuation parameter (t1 = 1);
  --   px1       previous value for solution vector v;
  --   px2       value for solution vector v corresponding to pt2;
  --   px3       value for solution vector v corresponding to pt3;
  --   fail      true if last correction was not successful.

  -- ON RETURN :
  --   t         predicted value for the continuation parameter t;
  --   v         predicted value for the solution vector v;
  --   step      is actual step size taken.

    d0,d1,d2 : double_float;

  begin
    t := t + Create(step);
    if REAL_PART(t) > 1.0
     then step := REAL_PART(t) - 1.0;
          t := t1;
    end if;
    if fail then
      put_line(file,"No predictor for solution.");
    else
      d0 := AbsVal(pt0 - pt1);
      if 1.0 + d0 = 1.0 then
        put_line(file,"No predictor for solution.");
      else
        d1 := AbsVal(pt1 - pt2);
        if 1.0 + d1 = 1.0 then
          put_line(file,"Secant predictor for solution.");
          d0 := step/d0;
          v := v + Create(d0)*(v - px1);
        else
          d2 := AbsVal(pt2 - pt3);
          if 1.0 + d2 = 1.0
           then put_line(file,"Quadratic predictor for solution.");
                v := Quadratic_Predictor(t,pt0,pt1,pt2,v,px1,px2);
           else put_line(file,"Cubic predictor for solution.");
                v := Cubic_Predictor(t,pt0,pt1,pt2,pt3,v,px1,px2,px3);
          end if;
        end if;
      end if;
    end if;
  end Predictor;

-- TARGET ROUTINES :

  procedure Silent_LU_Track 
               ( f : in Eval_Poly_Sys; jf : in Eval_Jaco_Mat;
                 s : in out Solu_Info;
                 pp : in Pred_Pars; cp : in Corr_Pars ) is

    pt0,pt1,pt2,pt3 : Complex_Number := s.sol.t;
    px0,px1,px2,px3 : Vector(s.sol.v'range) := s.sol.v;
    step : double_float := pp.maxstep;
    plane : Matrix(jf'range(2),0..s.sol.v'last);
    fail : boolean := true;
    nbit,nbsuccess : natural := 0;
    t1 : Complex_Number := Create(1.0);

  begin
    while not At_End(s,t1,cp.maxtot) or fail loop
      Predictor(s.sol.t,s.sol.v,step,pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
      plane := Path(s.sol.t);
      Affine_LU_Newton(f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                       s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      s.niter := s.niter + nbit;
      s.nstep := s.nstep + 1;
      if fail then s.nfail := s.nfail + 1; end if;
      Step_Control(fail,pp,step,nbsuccess,s.sol.t,
                   pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      exit when (step <= pp.minstep);
    end loop;
    if (step > pp.minstep)
     then plane := Path(Create(1.0));
          Affine_LU_Newton
            (f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
             s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,s.rcond,fail);
    end if;
  end Silent_LU_Track;

  procedure Reporting_LU_Track
               ( file : in file_type; 
                 f : in Eval_Poly_Sys; jf : in Eval_Jaco_Mat;
                 s : in out Solu_Info;
                 pp : in Pred_Pars; cp : in Corr_Pars ) is

    pt0,pt1,pt2,pt3 : Complex_Number := s.sol.t;
    px0,px1,px2,px3 : Vector(s.sol.v'range) := s.sol.v;
    step : double_float := pp.maxstep;
    plane : Matrix(jf'range(2),0..s.sol.v'last);
    fail : boolean := true;
    nbit,nbsuccess : natural := 0;
    t1 : Complex_Number := Create(1.0);
    oc : output_code := Process_io.Get_Output_Code;

  begin
    while not At_End(s,t1,cp.maxtot) or fail loop
      if (oc = p) or (oc = sp) or (oc = pc) or (oc = spc) then
        Predictor(file,s.sol.t,s.sol.v,step,
                  pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
        Process_io.pWrite(file,s.nstep,step,s.sol.t);
      else
        Predictor(s.sol.t,s.sol.v,step,pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
      end if;
      plane := Path(s.sol.t);
      if (oc = c) or (oc = sc) or (oc = pc) or (oc = spc) then
        Affine_LU_Newton
          (file,f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
           s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      else
        Affine_LU_Newton
          (f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
           s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      end if;
      s.niter := s.niter + nbit;
      s.nstep := s.nstep + 1;
      if fail then s.nfail := s.nfail + 1; end if;
      if (oc = p) or (oc = sp) or (oc = pc) or (oc = spc) then
        Step_Control(file,fail,pp,step,nbsuccess,s.sol.t,
                     pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      else
        Step_Control(fail,pp,step,nbsuccess,s.sol.t,
                     pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      end if;
      exit when (step <= pp.minstep);
    end loop;
    if (step > pp.minstep) then
      plane := Path(Create(1.0));
      if (oc = c) or (oc = sc) or (oc = pc) or (oc = spc) then
        Affine_LU_Newton
          (file,f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
           s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,s.rcond,fail);
      else
        Affine_LU_Newton
          (f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
           s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,s.rcond,fail);
      end if;
    end if;
  end Reporting_LU_Track;

  procedure Silent_QR_Track
               ( f : in Eval_Poly_Sys; jf : in Eval_Jaco_Mat;
                 s : in out Solu_Info;
                 pp : in Pred_Pars; cp : in Corr_Pars ) is

    pt0,pt1,pt2,pt3 : Complex_Number := s.sol.t;
    px0,px1,px2,px3 : Vector(s.sol.v'range) := s.sol.v;
    step : double_float := pp.maxstep;
    plane : Matrix(jf'range(2),0..s.sol.v'last);
    fail : boolean := true;
    nbit,nbsuccess : natural := 0;
    t1 : Complex_Number := Create(1.0);

  begin
    while not At_End(s,t1,cp.maxtot) or fail loop
      Predictor(s.sol.t,s.sol.v,step,pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
      plane := Path(s.sol.t);
      Affine_QR_Newton(f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                       s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      s.niter := s.niter + nbit;
      s.nstep := s.nstep + 1;
      if fail then s.nfail := s.nfail + 1; end if;
      Step_Control(fail,pp,step,nbsuccess,s.sol.t,
                   pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      exit when (step <= pp.minstep);
    end loop;
    if (step > pp.minstep)
     then plane := Path(Create(1.0));
          declare
            mm : constant natural := Min0(f'last+1,plane'last);
            sv : Vector(1..mm);
          begin
            Affine_SV_Newton
              (f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
               s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,sv,fail);
            s.rcond := Radius(sv(s.sol.v'last)/sv(sv'first));
          end;
    end if;
  end Silent_QR_Track;

  procedure Reporting_QR_Track
               ( file : in file_type; 
                 f : in Eval_Poly_Sys; jf : in Eval_Jaco_Mat;
                 s : in out Solu_Info;
                 pp : in Pred_Pars; cp : in Corr_Pars ) is

    pt0,pt1,pt2,pt3 : Complex_Number := s.sol.t;
    px0,px1,px2,px3 : Vector(s.sol.v'range) := s.sol.v;
    step : double_float := pp.maxstep;
    plane : Matrix(jf'range(2),0..s.sol.v'last);
    fail : boolean := true;
    nbit,nbsuccess : natural := 0;
    t1 : Complex_Number := Create(1.0);
    oc : output_code := Process_io.Get_Output_Code;

  begin
    while not At_End(s,t1,cp.maxtot) or fail loop
      if (oc = p) or (oc = sp) or (oc = pc) or (oc = spc) then
        Predictor(file,s.sol.t,s.sol.v,step,
                  pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
        Process_io.pWrite(file,s.nstep,step,s.sol.t);
      else
        Predictor(s.sol.t,s.sol.v,step,pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
      end if;
      plane := Path(s.sol.t);
      if (oc = c) or (oc = sc) or (oc = pc) or (oc = spc) then
        Affine_QR_Newton
          (file,f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
           s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      else
        Affine_QR_Newton
          (f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
           s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      end if;
      s.niter := s.niter + nbit;
      s.nstep := s.nstep + 1;
      if fail then s.nfail := s.nfail + 1; end if;
      if (oc = p) or (oc = sp) or (oc = pc) or (oc = spc) then
        Step_Control(file,fail,pp,step,nbsuccess,s.sol.t,
                     pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      else
        Step_Control(fail,pp,step,nbsuccess,s.sol.t,
                     pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      end if;
      exit when (step <= pp.minstep);
    end loop;
    if (step > pp.minstep) then
      plane := Path(Create(1.0));
      declare
        mm : constant natural := Min0(f'last+1,plane'last);
        sv : Vector(1..mm);
      begin
        if (oc = c) or (oc = sc) or (oc = pc) or (oc = spc) then
          Affine_SV_Newton
            (file,f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
             s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,sv,fail);
        else
          Affine_SV_Newton
            (f,jf,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
             s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,sv,fail);
        end if;
        s.rcond := Radius(sv(s.sol.v'last)/sv(sv'first));
      end;
    end if;
  end Reporting_QR_Track;

-- GENERIC TARGET ROUTINES :

  procedure G_Silent_LU_Track 
               ( ne,nv : in natural; s : in out Solu_Info;
                 pp : in Pred_Pars; cp : in Corr_Pars ) is

    pt0,pt1,pt2,pt3 : Complex_Number := s.sol.t;
    px0,px1,px2,px3 : Vector(s.sol.v'range) := s.sol.v;
    step : double_float := pp.maxstep;
    plane : Matrix(1..nv,0..s.sol.v'last);
    fail : boolean := true;
    nbit,nbsuccess : natural := 0;
    t1 : Complex_Number := Create(1.0);

    procedure Newton is new Silent_Affine_LU_Newton(f,jf);
    procedure RCO_Newton is new Silent_Affine_LU_RCO_Newton(f,jf);

  begin
    while not At_End(s,t1,cp.maxtot) or fail loop
      Predictor(s.sol.t,s.sol.v,step,pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
      plane := Path(s.sol.t);
      Newton(ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
             s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      s.niter := s.niter + nbit;
      s.nstep := s.nstep + 1;
      if fail then s.nfail := s.nfail + 1; end if;
      Step_Control(fail,pp,step,nbsuccess,s.sol.t,
                   pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      exit when (step <= pp.minstep);
    end loop;
    if (step > pp.minstep)
     then plane := Path(Create(1.0));
          RCO_Newton(ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                     s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,s.rcond,fail);
    end if;
  end G_Silent_LU_Track;

  procedure G_Reporting_LU_Track
               ( file : in file_type; 
                 ne,nv : in natural; s : in out Solu_Info;
                 pp : in Pred_Pars; cp : in Corr_Pars ) is

    pt0,pt1,pt2,pt3 : Complex_Number := s.sol.t;
    px0,px1,px2,px3 : Vector(s.sol.v'range) := s.sol.v;
    step : double_float := pp.maxstep;
    plane : Matrix(1..nv,0..s.sol.v'last);
    fail : boolean := true;
    nbit,nbsuccess : natural := 0;
    t1 : Complex_Number := Create(1.0);
    oc : output_code := Process_io.Get_Output_Code;

    procedure R_Newton is new Reporting_Affine_LU_Newton(f,jf);
    procedure S_Newton is new Silent_Affine_LU_Newton(f,jf);
    procedure R_RCO_Newton is new Reporting_Affine_LU_RCO_Newton(f,jf);
    procedure S_RCO_Newton is new Silent_Affine_LU_RCO_Newton(f,jf);

  begin
    while not At_End(s,t1,cp.maxtot) or fail loop
      if (oc = p) or (oc = sp) or (oc = pc) or (oc = spc) then
        Predictor(file,s.sol.t,s.sol.v,step,
                  pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
        Process_io.pWrite(file,s.nstep,step,s.sol.t);
      else
        Predictor(s.sol.t,s.sol.v,step,pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
      end if;
      plane := Path(s.sol.t);
      if (oc = c) or (oc = sc) or (oc = pc) or (oc = spc) then
        R_Newton(file,ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                 s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      else
        S_Newton(ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                 s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      end if;
      s.niter := s.niter + nbit;
      s.nstep := s.nstep + 1;
      if fail then s.nfail := s.nfail + 1; end if;
      if (oc = p) or (oc = sp) or (oc = pc) or (oc = spc) then
        Step_Control(file,fail,pp,step,nbsuccess,s.sol.t,
                     pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      else
        Step_Control(fail,pp,step,nbsuccess,s.sol.t,
                     pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      end if;
      exit when (step <= pp.minstep);
    end loop;
    if (step > pp.minstep) then
      plane := Path(Create(1.0));
      if (oc = c) or (oc = sc) or (oc = pc) or (oc = spc) then
        R_RCO_Newton(file,ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                     s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,s.rcond,fail);
      else
        S_RCO_Newton(ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                     s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,s.rcond,fail);
      end if;
    end if;
  end G_Reporting_LU_Track;

  procedure G_Silent_QR_Track
               ( ne,nv : in natural; s : in out Solu_Info;
                 pp : in Pred_Pars; cp : in Corr_Pars ) is

    pt0,pt1,pt2,pt3 : Complex_Number := s.sol.t;
    px0,px1,px2,px3 : Vector(s.sol.v'range) := s.sol.v;
    step : double_float := pp.maxstep;
    plane : Matrix(1..nv,0..s.sol.v'last);
    fail : boolean := true;
    nbit,nbsuccess : natural := 0;
    t1 : Complex_Number := Create(1.0);

    procedure QR_Newton is new Silent_Affine_QR_Newton(f,jf);
    procedure SV_Newton is new Silent_Affine_SV_Newton(f,jf);

  begin
    while not At_End(s,t1,cp.maxtot) or fail loop
      Predictor(s.sol.t,s.sol.v,step,pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
      plane := Path(s.sol.t);
      QR_Newton(ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      s.niter := s.niter + nbit;
      s.nstep := s.nstep + 1;
      if fail then s.nfail := s.nfail + 1; end if;
      Step_Control(fail,pp,step,nbsuccess,s.sol.t,
                   pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      exit when (step <= pp.minstep);
    end loop;
    if (step > pp.minstep)
     then plane := Path(Create(1.0));
          declare
            mm : constant natural := Min0(ne+1,plane'last);
            sv : Vector(1..mm);
          begin
            SV_Newton(ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                      s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,sv,fail);
            s.rcond := Radius(sv(s.sol.v'last)/sv(sv'first));
          end;
    end if;
  end G_Silent_QR_Track;

  procedure G_Reporting_QR_Track
               ( file : in file_type; 
                 ne,nv : in natural; s : in out Solu_Info;
                 pp : in Pred_Pars; cp : in Corr_Pars ) is

    pt0,pt1,pt2,pt3 : Complex_Number := s.sol.t;
    px0,px1,px2,px3 : Vector(s.sol.v'range) := s.sol.v;
    step : double_float := pp.maxstep;
    plane : Matrix(1..nv,0..s.sol.v'last);
    fail : boolean := true;
    nbit,nbsuccess : natural := 0;
    t1 : Complex_Number := Create(1.0);
    oc : output_code := Process_io.Get_Output_Code;

    procedure R_QR_Newton is new Reporting_Affine_QR_Newton(f,jf);
    procedure S_QR_Newton is new Silent_Affine_QR_Newton(f,jf);
    procedure R_SV_Newton is new Reporting_Affine_SV_Newton(f,jf);
    procedure S_SV_Newton is new Silent_Affine_SV_Newton(f,jf);

  begin
    while not At_End(s,t1,cp.maxtot) or fail loop
      if (oc = p) or (oc = sp) or (oc = pc) or (oc = spc) then
        Predictor(file,s.sol.t,s.sol.v,step,
                  pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
        Process_io.pWrite(file,s.nstep,step,s.sol.t);
      else
        Predictor(s.sol.t,s.sol.v,step,pt0,pt1,pt2,pt3,t1,px1,px2,px3,fail);
      end if;
      plane := Path(s.sol.t);
      if (oc = c) or (oc = sc) or (oc = pc) or (oc = spc) then
        R_QR_Newton(file,ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                    s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      else
        S_QR_Newton(ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                    s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,fail);
      end if;
      s.niter := s.niter + nbit;
      s.nstep := s.nstep + 1;
      if fail then s.nfail := s.nfail + 1; end if;
      if (oc = p) or (oc = sp) or (oc = pc) or (oc = spc) then
        Step_Control(file,fail,pp,step,nbsuccess,s.sol.t,
                     pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      else
        Step_Control(fail,pp,step,nbsuccess,s.sol.t,
                     pt0,pt1,pt2,pt3,s.sol.v,px0,px1,px2,px3);
      end if;
      exit when (step <= pp.minstep);
    end loop;
    if (step > pp.minstep) then
      plane := Path(Create(1.0));
      declare
        mm : constant natural := Min0(ne+1,plane'last);
        sv : Vector(1..mm);
      begin
        if (oc = c) or (oc = sc) or (oc = pc) or (oc = spc) then
          R_SV_Newton
            (file,ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
             s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,sv,fail);
        else
          S_SV_Newton(ne,plane,s.sol.v,cp.epsax,cp.epsrx,cp.epsaf,cp.epsrf,
                      s.cora,s.corr,s.resa,s.resr,nbit,cp.maxit,sv,fail);
        end if;
        s.rcond := Radius(sv(s.sol.v'last)/sv(sv'first));
      end;
    end if;
  end G_Reporting_QR_Track;

end Standard_Intrinsic_Trackers;
