with Standard_Complex_Numbers;           use Standard_Complex_Numbers;
with Standard_Complex_Vectors;           use Standard_Complex_Vectors;
with Standard_Complex_VecVecs;           use Standard_Complex_VecVecs;
with Multprec_Complex_VecVecs;
with Standard_Complex_Polynomials;       use Standard_Complex_Polynomials;
with Standard_Complex_Poly_Systems;      use Standard_Complex_Poly_Systems;
with Standard_Complex_Solutions;         use Standard_Complex_Solutions;

package Witness_Points is

-- DESCRIPTION :
--   To compute generic points on components of solutions, we add
--   a random hyperplane to the system and apply an embedding with
--   an additional variable to deal with the overdeterminedness.

-- GENERATE RANDOM SLICES :

  function Random_Hyperplanes ( k,n : natural )
                              return Standard_Complex_VecVecs.VecVec;
  function Random_Hyperplanes ( k,n,size : natural )
                              return Multprec_Complex_VecVecs.VecVec;

  -- DESCRIPTION :
  --   Returns k random hyperplanes, represented as vectors of range 0..n,
  --   either with coefficients as standard complex numbers,
  --   or with multi-precision complex numbers of the given size.

-- MANIPULATION OF POLYNOMIALS :

  function Add_Embedding ( p : Poly; k : natural ) return Poly;
  function Add_Embedding ( p : Poly_Sys; k : natural ) return Poly_Sys;

  -- DESCRIPTION :
  --   Does Add_Variables(p,k) and adds to the result k linear terms
  --   with random coefficients in the k extra variables.
  --   On a system p, the system on return has the same range.

-- OPERATIONS ON POLYNOMIAL SYSTEMS :

  function Add_Slice ( p : Poly_Sys; hyp : Vector ) return Poly_Sys;

  -- DESCRIPTION :
  --   Adds the hyperplane to the polynomial system p.  The coefficients of
  --   the hyperplane are indexed from 0 up to n = p'last, to represent
  --     hyp(0) + hyp(1)*x(1) + hyp(2)*x(2) + .. + hyp(n)*x(n) = 0, 
  --   The last equation of the system on return is the representation
  --   of the hyperplane as a polynomial.
  --   The equations of p are shared with those of the system on return.

  function Remove_Slice ( p : Poly_Sys ) return Poly_Sys;

  -- DESCRIPTION :
  --   Replaces the last equation of p by x_n = 0, where n = p'last. 

  function Eliminate_Slice ( p : Poly_Sys; k,i : natural ) return Poly_Sys;

  -- DESCRIPTION :
  --   Returns the polynomial system where the i-th unknown has been
  --   eliminated, using the k-th slice in the equations of p.

  -- REQUIRED : Degree(p(k),i) = 1.

  function Embed ( p : Poly_Sys ) return Poly_Sys;

  -- DESCRIPTION :
  --   Augments the number of variables in the polynomials with one,
  --   every monomial is multiplied with z^0.

  function Embed ( p : Poly_Sys; gamma : Vector ) return Poly_Sys;

  -- DESCRIPTION :
  --   To every polynomial in p the term gamma(i)*z is added, where
  --   z is a new variable.

  function Slice_and_Embed ( p : Poly_Sys; k : natural ) return Poly_Sys;

  -- DESCRIPTION :
  --   Adds k slices and k slack variables to the system p.
  -- REQUIRED :
  --   p is a square system.

  function Embed_with_Dummies ( p : Poly_Sys; k : natural ) return Poly_Sys;

  -- DESCRIPTION :
  --   Adds k dummy slack variables with k extra equations setting those
  --   dummy variables to zero.  The original last k hyperplanes of p
  --   (see the requirement) are replaced with the equations that set
  --   the dummy variables to zero.

  -- REQUIRED :
  --   p is a square system and contains at the end at least k hyperplanes.

  function Slice_and_Embed ( p : Poly_Sys; k : natural ) 
                           return Array_of_Poly_Sys;

  -- DESCRIPTION :
  --   Returns a sequence of embedded systems of range 0..k.
  --   The 0-th entry is the original system p, the i-th entry is obtained
  --   by adding i random hyperplanes and i additional variables to p.
  --   If k > 1, then the slices are orthogonal with respect to each other.

  function Slice_and_Embed ( p : Poly_Sys; k,l : natural )
                           return Array_of_Poly_Sys;

  -- DESCRIPTION :
  --   In constructing the sequence of embedded systems, like above,
  --   the last l equations of p are copied and used as slices.
  --   This is useful in the treatment of underdetermined systems.

  -- REQUIRED : l <= k.

  function Slices ( p : Poly_Sys; k : natural ) return VecVec;

  -- DESCRIPTION :
  --   Returns the vector representation of the k slices added to p.
  --   Each vector in the vector of vectors of return contains the
  --   coefficients cff:  cff(0) + cff(1)*x1 + .. + cff(n)*xn.

  function Square ( p : Poly_Sys ) return Poly_Sys;

  -- DESCRIPTION :
  --   The system on return is square, has as many equations as unknowns.
  --   If p is square, then Square(p) = p, otherwise slices or additional
  --   unknowns are added depending on whether p is has too few equations
  --   or too few unknowns, using the routines "Add_Slice" or "Embed".

  function Remove_Embedding ( p : Poly; dim : natural ) return Poly;
  function Remove_Embedding ( p : Poly_Sys; dim : natural ) return Poly_Sys;

  -- DESCRIPTION :
  --   Removes all dim added variables and the terms added to p in
  --   the embedding.

  function Complete ( n,k : natural; p : Poly_Sys ) return Poly_Sys;

  -- DESCRIPTION :
  --   When p is not a complete intersection, random multiples of the
  --   extra polynomials are added to have n-k equations on return.

  -- ON ENTRY :
  --   n         number of variables;
  --   k         dimension of the solution component;
  --   p         polynomial system.

  -- ON RETURN :
  --   n-k equations which also define the solution component.

-- OPERATIONS ON SOLUTION LISTS :

  function Add_Component ( s : Solution; c : Complex_Number ) return Solution;
 
  -- DESCRIPTION :
  --   Add c as last component to the solution vector.

  function Add_Component ( sols : Solution_List; c : Complex_Number )
                         return Solution_List;
 
  -- DESCRIPTION :
  --   Add c as last component to the solution vectors in the list.

  function Add_Embedding ( s : Solution; k : natural ) return Solution;
  function Add_Embedding ( sols : Solution_List; k : natural )
                         return Solution_List;

  -- DESCRIPTION :
  --   Adds k zeros to every solution.

  function Remove_Component ( s : Solution ) return Solution;
 
  -- DESCRIPTION :
  --   Removes the last component from the solution vector.

  procedure Remove_Component ( sols : in out Solution_List );
  function  Remove_Component ( sols : Solution_List ) return Solution_List;
 
  -- DESCRIPTION :
  --   Removes the last component from the solution vectors in the list.

  function Remove_Embedding ( s : Solution; k : natural ) return Solution;
  function Remove_Embedding ( sols : Solution_List; k : natural )
                            return Solution_List;

  -- DESCRIPTION :
  --   Removes the last k components from the solution.

end Witness_Points;
