with text_io;                           use text_io;
with Standard_Complex_Numbers;          use Standard_Complex_Numbers;
with Standard_Natural_Vectors;
with Standard_Complex_Vectors;
with Standard_Complex_Poly_Systems;     use Standard_Complex_Poly_Systems;
with Standard_Complex_Solutions;        use Standard_Complex_Solutions;

package Parameter_Homotopy_Continuation is

-- DESCRIPTION :
--   Parameter continuation assumes a system with strictly fewer equations
--   than unknowns.  The definition of start and target values for the
--   parameters lead to naturally to interpolation between the two.
--   This interpolation may be seen as the default function to instantiate
--   the general parameter continuation routine with.

  function Define_Start ( v : Standard_Complex_Vectors.Vector;
                          ip : Standard_Natural_Vectors.Vector )
                        return Standard_Complex_Vectors.Vector;

  -- DESCRIPTION :
  --   Returns the elements in v indexed by the parameters in ip.

  function Define_Target ( ip : Standard_Natural_Vectors.Vector )
                         return Standard_Complex_Vectors.Vector;

  -- DESCRIPTION :
  --   Given the indices to the parameters, this function prompts the
  --   user for target values of the parameters.

  procedure Write_Parameter_Values
              ( file : in file_type;
                labels : in Standard_Natural_Vectors.Vector;
                values : in Standard_Complex_Vectors.Vector );

  -- DESCRIPTION :
  --   Writes the values for the parameters in a nice format to file.

  function Interpolate ( a,b : Standard_Complex_Vectors.Vector;
                         t : Complex_Number )
                       return Standard_Complex_Vectors.Vector;

  -- DESCRIPTION :
  --   Returns (1-t)*a + t*b, for t between 0 and 1.  Is the 1st
  --   default to use in the instantiation of Parameter_Continuation.

  function Circulate ( a,b : Standard_Complex_Vectors.Vector;
                       gamma,t : Complex_Number )
                     return Standard_Complex_Vectors.Vector;

  -- DESCRIPTION :
  --   Returns (1-s)*a + s*b, for s = t + gamma*t*(1-t).  This move
  --   through the parameter space ensures complex arithmetic.

  function Differentiate ( a,b : Standard_Complex_Vectors.Vector )
                         return Standard_Complex_Vectors.Vector;

  -- DESCRIPTION :
  --   Just returns b-a as this is the derivative of the interpolate
  --   function of the parameters.  Is the 2nd default to use in the
  --   instantiation of Parameter_Continuation.

  generic
    with function Evaluate_Parameters ( t : Complex_Number )
           return Standard_Complex_Vectors.Vector;
    -- returns value of parameters at t
    with function Differentiate_Parameters ( t : Complex_Number )
           return Standard_Complex_Vectors.Vector;
    -- returns derivatives of parameters with respect to t
  procedure Parameter_Continuation
              ( file : in file_type;
                n : in natural; p : in Poly_Sys;
                pars : in Standard_Natural_Vectors.Vector;
                vars : in Standard_Natural_Vectors.Vector;
		sols : in out Solution_List; output : in boolean );

  -- DESCRIPTION :
  --   Executes the parameter continuation for the homotopy defined by
  --   the system p and for parameters defined by the generics functions.

  -- ON ENTRY :
  --   file     for intermediate output and diagnostics;
  --   n        number of unknowns and parameters in p.
  --   p        polynomial system with parameters;
  --   pars     indices to the parameter variables in p;
  --   vars     indices to the unknown variables in p;
  --   sols     start solutions for start values of parameters;
  --   output   true if intermediate output during continuation.

  -- ON RETURN :
  --   sols     solutions for target values of parameters.

  procedure Coefficient_Parameter_Homotopy_Continuation
              ( file : in file_type;
                p : in Poly_Sys; sols : in Solution_List;
                nb_equ,nb_unk,nb_par : in natural );

  -- DESCRIPTION :
  --   Interactive set up a coefficient-parameter homotopy 
  --   and calls then the path trackers.
  --   This is the main driver of this package.

  -- REQUIRED :
  --   All solutions have the same values for the parameters.

  -- ON ENTRY :
  --   file     file created for output;
  --   p        a polynomial system;
  --   sols     solutions for particular values of the parameters;
  --   nb_equ   number of equations in the system;
  --   nb_unk   number of unknowns in the system;
  --   nb_par   number of parameters: nb_unk - nb_equ.

end Parameter_Homotopy_Continuation;
