with text_io;                           use text_io;
with Standard_Natural_Vectors;
with Standard_Complex_Vectors;
with Standard_Complex_Polynomials;      use Standard_Complex_Polynomials;
with Standard_Complex_Poly_Systems;     use Standard_Complex_Poly_Systems;
with Standard_Complex_Solutions;        use Standard_Complex_Solutions;

package Systems_with_Parameters is

-- DESCRIPTION :
--   Any system with parameters is a coefficient-parameter homotopy.
--   Operations specific to such systems concern the definition of
--   the parameters, as interactively determined by the user.
--   Once the system has been solved for particular values of the
--   parameters, we need to substitute the parameters by these values
--   for root refinement.

  procedure Read_Solution_Parameters
              ( infile : in file_type; outfile : out file_type;
                p : in Poly_Sys; sols : out Solution_List;
                nb_equ,nb_unk,nb_par : out natural );

  -- DESCRIPTION :
  --   Scans the infile for the solutions for the system p.
  --   This procedure is called by Read_Parameter_Homotopy.

  -- REQUIRED : number of variables > number of equations

  -- ON ENTRY :
  --   infile   file with system already read from;
  --   p        a polynomial system with parameters.

  -- ON RETURN :
  --   sols     solutions for particular values of the parameters;
  --   nb_equ   number of equations in the system;
  --   nb_unk   number of unknowns in the system;
  --   nb_par   number of parameters: nb_unk - nb_equ.

  procedure Read_Parameter_Homotopy
              ( outfile : out file_type;
                lp : out Link_to_Poly_Sys; sols : out Solution_List;
                nb_equ,nb_unk,nb_par : out natural );

  -- DESCRIPTION :
  --   Reads a coefficient-paremeter homotopy with solutions for 
  --   some values of the parameters.

  -- ON RETURN :
  --   outfile  file created for output;
  --   lp       a polynomial system;
  --   sols     solutions for particular values of the parameters;
  --   nb_equ   number of equations in the system;
  --   nb_unk   number of unknowns in the system;
  --   nb_par   number of parameters: nb_unk - nb_equ.

  function Define_Parameters ( nb_equ,nb_unk,nb_par : natural )
                             return Standard_Natural_Vectors.Vector;

  -- DESCRIPTION :
  --   Returns the vector with indices to the parameters,
  --   as defined interactively by the user.
  --   The vector on return is sorted in increasing order.

  function Complement ( n : natural; v : Standard_Natural_Vectors.Vector )
                      return Standard_Natural_Vectors.Vector;

  -- DESCRIPTION :
  --   Returns the complement of the elements in v with respect to
  --   the set of natural numbers 1,2,..,n.

  -- NOTE: the complement of the parameters are those variables which
  --   depend on the natural parameters via the equations of the system.

  -- REQUIRED : the entries in v are sorted in increasing order.

  function Substitute
              ( t : Term;
                pars : Standard_Natural_Vectors.Vector;
                vals : Standard_Complex_Vectors.Vector ) return Term;

  -- DESCRIPTION :
  --   Replaces the parameters in t by the given values in vals.
  --   The parameters are defined by the indices in pars.

  function Substitute
              ( p : Poly;
                pars : Standard_Natural_Vectors.Vector;
                vals : Standard_Complex_Vectors.Vector ) return Poly;

  function Substitute
               ( p : Poly_Sys;
                 pars : Standard_Natural_Vectors.Vector;
                 vals : Standard_Complex_Vectors.Vector )
               return Poly_Sys;

  -- DESCRIPTION :
  --   Replaces the parameters in p by given values.

end Systems_with_Parameters;
