with Standard_Integer_Vectors;
with Standard_Floating_Vectors;
with Arrays_of_Floating_Vector_Lists;   use Arrays_of_Floating_Vector_Lists;
with Floating_Mixed_Subdivisions;       use Floating_Mixed_Subdivisions;

package Cells_Container is

-- DESCRIPTION :
--   This package provides a container for mixed-cell configurations,
--   designed for the interface with C.

-- CREATORS :

  procedure Initialize 
              ( mixture : in Standard_Integer_Vectors.Link_to_Vector;
                lifting : in Link_to_Array_of_Lists;
                mcc : in Mixed_Subdivision );

  -- DESCRIPTION :
  --   Initializes the container with a mixed-cell configuration.
  --   On entry are the type of mixture, lifted supports and the cells.

  procedure Initialize
              ( mixture : in Standard_Integer_Vectors.Link_to_Vector );

  -- DESCRIPTION :
  --   Initializes the container with a type of mixture.

  procedure Initialize ( lifting : in Link_to_Array_of_Lists );

  -- DESCRIPTION :
  --   Initializes the container with lifted supports.

  procedure Initialize ( mcc : in Mixed_Subdivision );

  -- DESCRIPTION :
  --   Initializes the container with a mixed-cell configuration.

-- SELECTORS :

  function Length return natural;

  -- DESCRIPTION :
  --   Returns the number of cells in the container.

  function Dimension return natural;

  -- DESCRIPTION :
  --   Returns 0 if the container is empty, otherwise the
  --   dimension of lifted points in the cells are returned.
  --   This dimension is the length of the lifted points.

  function Type_of_Mixture return Standard_Integer_Vectors.Link_to_Vector;

  -- DESCRIPTION :
  --   Returns the type of mixture of the mixed-cell configuration.

  function Lifted_Supports return Link_to_Array_of_Lists;

  -- DESCRIPTION :
  --   Returns the lifted supports for the mixed-cell configuration.

  procedure Retrieve ( k : in natural; mic : out Mixed_Cell;
                       fail : out boolean );

  -- DESCRIPTION : 
  --   Returns the k-th cell in the subdivision in mic if not fail.

  procedure Retrieve_Mixed_Cell
             ( k : in natural; fail : out boolean;
               cnt,lab : out Standard_Integer_Vectors.Link_to_Vector;
               normal : out Standard_Floating_Vectors.Link_to_Vector );

  -- DESCRIPTION :
  --   Returns the k-th cell in the subdivision if not fail,
  --   in a format most suitable for the C gateway interface.
  --   See the Append_Mixed_Cell below.

  -- ON ENTRY :
  --   k       index number to a mixed cell in the subdivision.

  -- ON RETURN :
  --   fail    true if there is no k-th cell;
  --   cnt     cnt(i) contains the number of points in the i-th support;
  --   lab     lab(i) is the label for the i-th points in the cell;
  --   normal  is the inner normal to the mixed cell.

  function Retrieve return Mixed_Subdivision;

  -- DESCRIPTION :
  --   Returns the mixed subdivision stored in the container.

-- CONSTRUCTORS :

  function Append_to_Support
             ( k : in natural;
               x : in Standard_Floating_Vectors.Vector ) return boolean;

  -- DESCRIPTION :
  --   Adds the point x to the k-th lifted support.
  --   The return value indicates failure and is true if k is out of
  --   the range of the type of mixture, otherwise false is returned.
  
  -- REQUIRED : the type of mixture is initialized.

  procedure Append ( mic : in Mixed_Cell );

  -- DESCRIPTION :
  --   Appends the cell to the current mixed-cell configuration.

  procedure Append_Mixed_Cell
             ( cnt,lab : in Standard_Integer_Vectors.Vector;
               normal : in Standard_Floating_Vectors.Vector );

  -- DESCRIPTION :
  --   Creates a mixed cell defined by the input parameters
  --   and appends this cell to the current mixed-cell configuration.
  --   See the Retrieve_Mixed_Cell above.

  -- ON ENTRY :
  --   cnt     cnt(i) contains the number of points in the i-th support;
  --   lab     lab(i) is the label for the i-th points in the cell;
  --   normal  is the inner normal to the mixed cell.

-- DESTRUCTOR :

  procedure Clear;

  -- DESCRIPTION :
  --   Clears the cells in the container.

end Cells_Container;
