with unchecked_deallocation;
with Standard_Floating_Numbers;         use Standard_Floating_Numbers;
with Standard_Floating_Numbers_io;      use Standard_Floating_Numbers_io;
with Standard_Complex_Numbers;          use Standard_Complex_Numbers;
with Standard_Complex_VecVecs;
with Standard_Complex_Norms_Equals;     use Standard_Complex_Norms_Equals;
with Standard_Complex_Poly_Systems;     use Standard_Complex_Poly_Systems;
with Standard_Complex_Solutions;        use Standard_Complex_Solutions;
with Sample_Point_Lists;                use Sample_Point_Lists;
with Rectangular_Sample_Grids;
with Monodromy_Component_Breakup;
with Systems_Container;
with Solutions_Container;
with Monodromy_Partitions;              use Monodromy_Partitions;
with Sampling_Operations;

with text_io,integer_io;                use text_io,integer_io;
with Standard_Complex_Numbers_io;       use Standard_Complex_Numbers_io;
with Standard_Complex_Vectors;
with Standard_Complex_VecVecs_io;       use Standard_Complex_VecVecs_io;
with Standard_Complex_Solutions_io;     use Standard_Complex_Solutions_io;
with Standard_Complex_VecVecs_io;

package body Monodromy_Permutations is

-- INTERNAL DATA :

  type Link_to_Array_of_Solution_Lists is access Array_of_Solution_Lists;

  ind : natural;
  dim : natural;
  grid : Link_to_Array_of_Solution_Lists;
  deco : Link_to_VecVec;
  traces_grid : Array_of_Standard_Sample_Lists(0..2);

-- OPERATIONS :

  procedure Initialize ( n,d,k : in natural ) is

    n2 : constant natural := n+2;

  begin
    ind := -1;
    dim := k;
    grid := new Array_of_Solution_Lists(0..n2);
    deco := Monodromy_Partitions.Init_Factors(d);
  end Initialize;

  procedure On_Slice ( sols : in Solution_List;
                       v : in Standard_Complex_Vectors.Vector ) is 

  -- DESCRIPTION :
  --   This routine prints the result of the evaluation of every
  --   solution at the hyperplane with coefficients in v.
  --   Good for checking purposes.
 
    tmp : Solution_List := sols;
    ls : Link_to_Solution;
    res : Complex_Number := Create(0.0);

  begin
    for i in 1..Length_Of(sols) loop
      ls := Head_Of(tmp);
      res := v(0);
      for j in v'range loop
        res := res + ls.v(j)*v(j);
      end loop;
      put("residual on slice : "); put(res); new_line;
      tmp := Tail_Of(tmp);
    end loop;
  end On_Slice;

  function Retrieve ( sols : Solution_List;
                      m : natural ) return Link_to_Solution is

  -- DESCRIPTION :
  --   Retrieves the solution with the m equal to the given m.

    tmp : Solution_List := sols;
    ls : Link_to_Solution;

  begin
    while not Is_Null(tmp) loop
      ls := Head_Of(tmp);
      if ls.m = m
       then return ls;
       else tmp := Tail_Of(tmp);
      end if;
    end loop;
    return ls;
  end Retrieve;

  function Sort ( sols : Solution_List ) return Solution_List is

  -- DESCRIPTION :
  --   Returns the same solution list, sorted in increasing order
  --   of the m field.  This is the needed for the trace grid.

    res,res_last : Solution_List;

  begin
    for i in 1..Length_Of(sols) loop
      Append(res,res_last,Retrieve(sols,i));
    end loop;
    return res;
  end Sort;

  procedure Store ( sols : in Solution_List ) is

    sli : Standard_Complex_VecVecs.Link_to_VecVec;
    err,dis : double_float;

  begin
    put("Storing with ind = "); put(ind,1); new_line;
    ind := ind + 1;
    Copy(sols,grid(ind));
   -- put(standard_output,Length_Of(sols),Head_Of(sols).n,sols);
    if ind < 3 then
      put("storing trace grid for slice "); put(ind,1); new_line;
      sli := Sampling_Operations.Retrieve_Start_Slices;
     -- put_line("the retrieved slices :"); put(sli);
     -- On_Slice(sols,sli(1).all);
      if ind = 0 then
        traces_grid(ind) := Create(sols,sli.all);
      else
        traces_grid(ind) := Create(Sort(sols),sli.all);
        if ind = 2 then
          err := Rectangular_Sample_Grids.Maximal_Error(traces_grid);
          put("error on trace grid : "); put(err,3); new_line;
          dis := Rectangular_Sample_Grids.Minimal_Distance(traces_grid);
          put("distance on trace grid : "); put(dis,3); new_line;
        end if;
      end if;
    end if;
  end Store;

  procedure Write_Grid is
  begin
    for i in 0..ind loop
      put("Solutions at grid("); put(i,1);
      put_line(") :");
      put(standard_output,Length_Of(grid(i)),Head_Of(grid(i)).n,grid(i));
    end loop;
  end Write_Grid;

  procedure Extinguish ( p : in out Vector ) is

  -- DESCRIPTION :
  --   For every i for which p(i) = 0, the corresponding j,
  --   for which p(j) = i will be set to zero as well.

    done : boolean;

  begin
    for i in p'range loop
      done := true;
      if p(i) = 0 then
        for j in p'range loop
          if p(j) = i then
            p(j) := 0;
            done := false;
          end if;
        end loop;
      end if;
      exit when done;
    end loop;
  end Extinguish;

  function Permutation return Vector is

    len : constant natural := Length_Of(grid(0));
    res : Vector(1..len);
    tol : constant double_float := 1.0E-8;
    pt1,pt2 : Solution_List;
    ls1,ls2 : Link_to_Solution;

  begin
   -- Write_Grid;
   -- put_line("Solution list 0 :");
   -- put(standard_output,Length_Of(grid(0)),Head_Of(grid(0)).n,grid(0));
   -- put("Solution list "); put(ind,1); put_line(" :");
   -- put(standard_output,Length_Of(grid(ind)),Head_Of(grid(ind)).n,grid(ind));
    pt1 := grid(0);
    for i in 1..len loop
      ls1 := Head_Of(pt1);
      if ls1.m > 0 then            -- ignore clustered and failed solutions
        pt2 := grid(ind);
        res(i) := 0;
        while not Is_Null(pt2) loop
          ls2 := Head_Of(pt2);
          if ls2.m > 0 then
            if Equal(ls1.v,ls2.v,tol)
             then res(i) := ls2.m;
            end if;
          end if;
          exit when (res(i) > 0);
          pt2 := Tail_Of(pt2);
        end loop;
      end if;
      pt1 := Tail_Of(pt1);
    end loop;
    Extinguish(res);
    return res;
  end Permutation;

  function Decomposition return Link_to_VecVec is
  begin
    return deco;
  end Decomposition;

  procedure Update_Decomposition
              ( p : in Vector; nf0,nf1 : out natural ) is

  begin
    if deco /= null
     then nf0 := Monodromy_Partitions.Number_of_Factors(deco.all);
          nf1 := nf0;
          Monodromy_Partitions.Add_Map(deco,nf1,p);
    end if;
  end Update_Decomposition;

  function Certify_with_Linear_Trace return boolean is

    done : boolean;
    tol : constant double_float := 1.0E-8;

  begin
    put_line("Certifying with linear trace test...");
    done := Monodromy_Component_Breakup.Is_Factorization
              (standard_output,tol,deco.all,traces_grid);
    return done;
  end Certify_with_Linear_Trace;

  procedure Clear is

    procedure free is 
      new unchecked_deallocation(Array_of_Solution_Lists,
                                 Link_to_Array_of_Solution_Lists);

  begin
    if grid /= null
     then for i in 0..ind loop
            Clear(grid(i));
          end loop;
          free(grid);
    end if;
    Standard_Natural_VecVecs.Deep_Clear(deco);
  end Clear;

end Monodromy_Permutations;
