/* simple test program in C on the Ada procedure use_solcon */

#include<stdio.h>
#include<stdlib.h>

extern void adainit();
extern int _ada_use_solcon ( int task, int *a, int *b, double *c );
extern void adafinal();

typedef struct list LIST;
struct list
{
   double rt;   /* real part of complex continuation parameter t */
   double it;   /* imaginary part of t */
   int m;       /* multiplicity of the solution vector */
   double *v;   /* solution is sequence of real and imaginary parts */
   double err;  /* error on the correction term from Newton's method */
   double rco;  /* inverse of estimate of condition number */
   double res;  /* residual of the solution vector */
   LIST *next;  /* next solution in the list */
};

LIST *push ( LIST *l, int n, double rt, double it, int m, double *v,
             double err, double rco, double res );
/* pushes the solution to the front of the list l */

void write_solution ( int n, double rt, double it, int m, double *v,
                      double err, double rco, double res );
/* writes the n-dimensional solution to the screen */

void write ( LIST *l, int n );
/* writes the list of n-dimensional solution vectors */

void test_solution_retrieval ( int n );
/* prompts the user for some solution number,
 * retrieves and prints the solution vector,
 * n is the complex dimension of the solution */

void test_replace_solution ( int n );
/* prompts the user for some solution number,
 * and replaces the solution */

LIST *test_retrievals ( int len, int n );
/* writes the solution list of length len and complex dimension n
 * on standard output, returns the list of solutions */

void test_additions ( int n, LIST *l );
/* fills up the container with the solutions in the list l */

int main(void)
{
   printf("\nTesting Solutions Container...\n");

   adainit();
   {
      int k,len,n,fail;
      double *c;
      LIST *l;

      fail = _ada_use_solcon(0,&k,&k,c);          /* read solution list */
      printf("The solution list :\n");
      fail = _ada_use_solcon(1,&k,&k,c);          /* print list to screen */
      fail = _ada_use_solcon(2,&k,&len,c);        /* get length of list */
      printf("Number of solutions : %d\n", len);
      /* get the dimension of the solutions in the container */
      fail = _ada_use_solcon(3,&k,&n,c);          /* get the dimension */
      printf("Dimension of vectors : %d\n", n);
      /* test retrievals of solutions */
      test_solution_retrieval(n);
      test_replace_solution(n);
      l = test_retrievals(len,n);
      printf("The solution list is reverse order : \n");
      write(l,n);
      fail = _ada_use_solcon(7,&k,&k,c);          /* clear the container */
      test_additions(n,l);                        /* reconstruct */
      printf("The reconstructed list in reverse order :\n");
      fail = _ada_use_solcon(1,&k,&k,c); 
   }
   adafinal();

   return 0;
}

void retrieve_and_write ( int n, int k )
{
   int i,m,fail;
   double sol[2*n+5];

   fail = _ada_use_solcon(4,&k,&m,sol);

   if (fail == 1)
      printf("some failure occurred...\n");
   else
   { 
      printf("The solution vector :\n");
      for(i=0; i<n; i++)
         printf(" %.15e  %.15e\n", sol[2+2*i], sol[2+2*i+1]);
   }
}

void test_solution_retrieval ( int n )
{
   int k;

   printf("Give k : "); scanf("%d", &k);

   retrieve_and_write(n,k);
}

void test_replace_solution ( int n )
{
   int i,k,m[2],fail;
   double sol[2*n+5];

   printf("Give k : "); scanf("%d", &k);
   retrieve_and_write(n,k);

   printf("Setting solution vector to 1,2,3,...\n");
   sol[0] = 1.111;
   sol[1] = 2.222;
   m[0] = n;
   m[1] = 12345;
   sol[2*n+2] = 3.333;
   sol[2*n+3] = 4.444;
   sol[2*n+4] = 5.555; 
   for(i=0; i<2*n; i++) sol[2+i] = i;

   fail = _ada_use_solcon(5,&k,m,sol);

   if (fail == 1)
      printf("some failure occurred...\n");
   else
   { 
      printf("The changed solution retrieved again :\n");
      retrieve_and_write(n,k);
   }
}

void write_solution ( int n, double rt, double it, int m, double *v,
                      double err, double rco, double res )
{
   int i;

   printf("t : %.15e  %.15e\n", rt, it);
   printf("m : %d\n", m);
   printf("The solution for t :\n");
   for(i=0; i<n; i++)
      printf(" %.15e  %.15e\n", v[2*i], v[2*i+1]);
   printf("== err : %.3e = rco : %.3e = res : %.3e ==\n",err,rco,res);
}

LIST *push ( LIST *l, int n, double rt, double it, int m, double *v,
             double err, double rco, double res )
{
   int i;
   LIST *nl = (LIST*)calloc(1,sizeof(LIST));

   nl->rt = rt;
   nl->it = it;
   nl->m = m;
   nl->v = calloc(2*n,sizeof(double));
   for(i=0; i<2*n; i++)
      nl->v[i] = v[i];
   nl->err = err;
   nl->rco = rco;
   nl->res = res;
   nl->next = l;

   return nl;
}

void write ( LIST *l, int n )
{
   LIST *p;
   int k;

   for(p=l,k=1; p!=NULL; p=p->next,k++)
   {
      printf("Solution %d :\n",k);
      write_solution(n,p->rt,p->it,p->m,p->v,p->err,p->rco,p->res);
   }
}

LIST *test_retrievals ( int len, int n )
{
   int i,k,m,fail;
   double sol[2*n+5],err,rco,res;
   LIST *l = NULL;

   for(k=1; k<=len; k++)
   {
      fail = _ada_use_solcon(4,&k,&m,sol);
      printf("Solution %d :\n", k);
      err = sol[2*n+2];
      rco = sol[2*n+3];
      res = sol[2*n+4];
      write_solution(n,sol[0],sol[1],m,sol+2,err,rco,res);
      l = push(l,n,sol[0],sol[1],m,sol+2,err,rco,res);
   }

   return l;
}

void test_additions ( int n, LIST *l )
{
   int i,k,m[2],fail;
   LIST *p;
   double sol[2*n+5];

   m[0] = n;
   for(p=l,k=1; p!=NULL; p=p->next,k++)
   {
      sol[0] = p->rt;
      sol[1] = p->it;
      m[1] = p->m;
      for(i=0;i<2*n;i++) sol[2+i] = p->v[i];
      sol[2*n+2] = p->err;
      sol[2*n+3] = p->rco;
      sol[2*n+4] = p->res;
      fail = _ada_use_solcon(6,&k,m,sol);
   }
}
