with text_io,integer_io;                use text_io,integer_io;
with Standard_Floating_Numbers;         use Standard_Floating_Numbers;
with Standard_Floating_Numbers_io;      use Standard_Floating_Numbers_io;
with Communications_with_User;          use Communications_with_User;
with Standard_Complex_Numbers;          use Standard_Complex_Numbers;
with Standard_Random_Numbers;           use Standard_Random_Numbers;
with Standard_Natural_Vectors;
with Standard_Natural_Vectors_io;       use Standard_Natural_Vectors_io;
with Standard_Complex_Vectors;
with Standard_Complex_VecVecs;
with Standard_Complex_Polynomials;      use Standard_Complex_Polynomials;
with Standard_Complex_Polynomials_io;   use Standard_Complex_Polynomials_io;
with Standard_Complex_Poly_Systems;
with Standard_Complex_Poly_Systems_io;  use Standard_Complex_Poly_Systems_io;
with Standard_Complex_Solutions;        use Standard_Complex_Solutions;
with Standard_Solution_Filters;         use Standard_Solution_Filters;
with Planes_and_Polynomials;            use Planes_and_Polynomials;
with Witness_Points,Witness_Points_io;  use Witness_Points,Witness_Points_io;
with Drivers_to_Cascade_Filtering;      use Drivers_to_Cascade_Filtering;
with Sampling_Machine;
with Sample_Points;                     use Sample_Points;
with Sample_Point_Lists;                use Sample_Point_Lists;
with Standard_Stacked_Sample_Grids;
with Drivers_to_Grid_Creators;          use Drivers_to_Grid_Creators;
with Standard_Trace_Interpolators;

procedure ts_dae is

-- DESCRIPTION :
--   Test on finding the index of a Differential Algebraic Equation.

  function Coefficients_of_Slice_in_J0
               ( dimj0,dimj1,dimsys : natural )
               return Standard_Complex_Vectors.Vector is

  -- DESCRIPTION :
  --   Returns the coefficients of a random slice parallel to J^0.

  -- ON ENTRY :
  --   dimj0     number of orginal variables in the space J^0;
  --   dimj1     dimension of the ambient space J^1;
  --   dimsys    dimension of the solution set of the system;

  -- ON RETURN :
  --   vector of length 0..dimj1 with random coefficients in range
  --   0..dimj0, dimj1-dymsys+1..dimj1, zero elsewhere.

    res : Standard_Complex_Vectors.Vector(0..dimj1+dimsys);

  begin
    for i in 0..dimj0 loop
      res(i) := Random1;
    end loop;
    for i in dimj0+1..dimj1 loop
      res(i) := Create(0.0);
    end loop;
    for i in dimj1+1..dimj1+dimsys loop
      res(i) := Random1;
    end loop;
    return res;
  end Coefficients_of_Slice_in_J0;

  function Embed ( p : Standard_Complex_Poly_Systems.Poly_Sys;
                   dimj0,dimj1,dimsys : natural )
                 return Standard_Complex_Poly_Systems.Poly_Sys is

    nv : constant natural := dimj1 + dimsys;
    res : Standard_Complex_Poly_Systems.Poly_Sys(1..nv);
    t : Term;

  begin
    t.dg := new Standard_Natural_Vectors.Vector'(1..nv => 0);
    for i in p'range loop
      res(i) := Add_Variables(p(i),dimsys);
      for j in 1..dimsys loop
        t.cf := Random1;
        t.dg(nv-j+1) := 1;
        Add(res(i),t);
        t.dg(nv-j+1) := 0;
      end loop;
    end loop;
    t.cf := Create(1.0);
    for i in 1..dimsys loop
      t.dg(nv-dimsys+i) := 1;
      res(p'last+i) := Create(t);
      t.dg(nv-dimsys+i) := 0;
    end loop;
    for i in dimj1+1..dimj1+dimsys loop
      declare
        slcff : Standard_Complex_Vectors.Vector(0..nv)
              := Coefficients_of_Slice_in_J0(dimj0,dimj1,dimsys);
        slice : Poly := Hyperplane(slcff);
      begin
        res(i) := slice;
      end;
    end loop;
    return res;
  end Embed;

  function fnormal ( p : in Standard_Complex_Polynomials.Poly;
                     tol : in double_float )
                   return Standard_Complex_Polynomials.Poly is

  -- DESCRIPTION :
  --   Only terms in p whose coefficient is larger than tol are returned.

    res : Poly := Null_Poly;

    procedure Normalize_Term ( t : in Term; continue : out boolean ) is
    begin
      if AbsVal(t.cf) > tol
       then Add(res,t);
      end if;
      continue := true;
    end Normalize_Term;
    procedure Normalize_Terms is new Visiting_Iterator(Normalize_Term);   

  begin
    Normalize_Terms(p);
    return res;
  end fnormal;

  procedure Decompose
               ( file : in file_type;
                 p : in Standard_Complex_Poly_Systems.Poly_Sys;
                 sols : in Solution_List; dim : in natural;
                 tol : in double_float;
                 elimp : out Standard_Complex_Polynomials.Poly;
                 maxerr : out double_float ) is

  -- DESCRIPTION :
  --   To decompose the solution set given by an embedded system p
  --   and the generic points in sols, we take the following steps :
  --    1) construct a grid of sample points;
  --    2) project and interpolate the sampled points; and
  --    3) expand the interpolating polynomial.

  -- ON ENTRY :
  --   file      to write intermedate output and diagnostics;
  --   p         embedded polynomial system;
  --   sols      generic points on random slices in J^0 space;
  --   dim       dimension of the solution set equals the number
  --             of slack variables;
  --   tol       tolerance to decide whether to neglect coefficient.

  -- ON RETURN :
  --   elimp     eliminated polynomial.

    use Standard_Stacked_Sample_Grids,Standard_Trace_Interpolators;
 
    sli : Standard_Complex_VecVecs.VecVec(1..dim) := Slices(p,dim);
    sps : Standard_Sample_List := Create(sols,sli);
    deg : constant natural := Length_Of(sps);
    ip,np : Standard_Complex_Polynomials.Poly;
 
  begin
    Sampling_Machine.Initialize(p);
    Sampling_Machine.Default_Tune_Sampler(2);
    Sampling_Machine.Default_Tune_Refiner;
    if dim = 1
     then declare
            grid : Array_of_Standard_Sample_Lists(0..deg);
            eps,dst : double_float;
            t : Trace_Interpolator1;
          begin
            Standard_Rectangular_Grid_Creator(file,sps,deg,grid,eps,dst);
            t := Create(grid);
            maxerr := Maximal_Error(t,grid);
            put(file,"Maximal residual of evaluation at the grid : ");
            put(file,maxerr,3); new_line(file);
            ip := Expand(t);
          end;
     else declare
            grid : Stacked_Sample_Grid(dim,deg);
            t : Trace_Interpolator;
          begin
            Standard_Stacked_Grid_Creator(file,sps,true,grid);
            t := Create(grid,deg);
            maxerr := Maximal_Error(t,grid);
            put(file,"Maximal residual of evaluation at the grid : ");
            put(file,maxerr,3); new_line(file);
            ip := Expand(t);
          end;
    end if;
    put_line(file,"The trace interpolator expanded as polynomial : ");
    put_line(file,ip);                                                          
    np := fnormal(ip,tol);
    put_line(file,"The normalized polynomial omitting small coefficients : ");
    put_line(file,np);
    elimp := np;
  end Decompose;

  procedure Eliminate
               ( file : in file_type;
                 p : in Standard_Complex_Poly_Systems.Poly_Sys;
                 dimj0,dimj1,dimsys,stage : in natural;
                 tol : in double_float;
                 constraint : out Standard_Complex_Polynomials.Poly;
                 fail : out boolean ) is

  -- DESCRIPTION :
  --   This procedure does the following steps :
  --     1) adds slices parallel to the J^0 space, embeds the system;
  --     2) solves the embedded polynomial system; and
  --     3) uses interpolation to find the polynomial.

  -- ON ENTRY :
  --   file      for intermediate output and results;
  --   p         polynomial system representing the DAE;
  --   dimj0     number of orginal variables in the space J^0;
  --   dimj1     dimension of the ambient space J^1;
  --   dimsys    dimension of the solution set of the system;
  --   stage     the number of the stage in geometric completion process;
  --   tol       tolerance to decide whether coefficient is zero.

  -- ON RETURN :
  --   constraint is the new constraint found by elimination;
  --   fail      is true when the maximal error of the interpolator
  --             on the grid is larger than the tolerance.

    dim : constant natural := dimsys-stage+1;
    nv : constant natural := dimj1 + dim;
    ep : Standard_Complex_Poly_Systems.Poly_Sys(1..nv);
    sols,gpsols : Solution_List;
    rc : natural;
    solvetime : duration;
    maxerr : double_float;

  begin
    new_line(file);
    put(file,"Eliminate at Stage "); put(file,stage,1); 
    put(file," for solution set of dimension ");
    put(file,dim,1); put_line(file," :");
    ep := Embed(p,dimj0,dimj1,dim);
    put_line(file,"The embedded polynomial system : ");
    put_line(file,ep);
    Black_Box_Solver(file,ep,true,sols,rc,solvetime);
    gpsols := Vanishing_Filter(sols,tol);
    Decompose(file,ep,gpsols,dim,tol,constraint,maxerr);
    fail := (maxerr > tol);
  end Eliminate;

  function Prolong ( dimj0,dimj1 : natural; p : Poly ) return Poly is
 
  -- DESCRIPTION :
  --   Given a polynomial whose independent variables depend on t,
  --   the derivative of p with the chain rule is returned.
 
    res : Poly := Null_Poly;
 
    procedure Prolong_Term ( t : in Term; cont : out boolean ) is
 
      dt : Term;
      fac : Complex_Number;
 
    begin
      dt.dg := new Standard_Natural_Vectors.Vector'(1..dimj1 => 0);
      dt.dg(t.dg'range) := t.dg.all;
      for i in 1..dimj0 loop
        if t.dg(i) > 0
         then fac := Create(t.dg(i));
              dt.cf := t.cf*fac;
              dt.dg(i) := t.dg(i)-1;
              dt.dg(i+dimj0) := 1;
              Add(res,dt);
              dt.dg(i) := t.dg(i);
              dt.dg(i+dimj0) := 0;
        end if;
      end loop;
      Clear(dt.dg);
      cont := true;
    end Prolong_Term;
    procedure Prolong_Terms is new Visiting_Iterator(Prolong_Term);
 
  begin
    Prolong_Terms(p);
    return res;
  end Prolong;

  procedure Differentiate
               ( file : in file_type;
                 constraint : in Standard_Complex_Polynomials.Poly;
                 dimj0,dimj1,i : in natural;
                 diffcons : out Standard_Complex_Polynomials.Poly ) is
  begin
    new_line(file);
    put(file,"Differentiate at Stage ");
    put(file,i,1); put_line(file," :");
    diffcons := Prolong(dimj0,dimj1,constraint);
    new_line(file);
    put_line(file,"The differentiated constraint : ");
    put_line(file,diffcons);
  end Differentiate;

  procedure Diff_Elim
               ( file : in file_type;
                 p : in Standard_Complex_Poly_Systems.Poly_Sys;
                 dimj0 : in natural ) is

  -- DESCRIPTION :
  --   Given the polynomial system that represents the DAE with
  --   dimj0 the dimension of the ambient space J^0.

    nbequ : constant natural := p'length;
    nbunk : constant natural := Number_of_Unknowns(p(p'first));
    dimsys : constant natural := nbunk - nbequ;
    tol : constant double_float := 1.0E-8;
    constraint,diffcons,newcons : Standard_Complex_Polynomials.Poly;
    wrk : Standard_Complex_Poly_Systems.Poly_Sys(1..nbunk+2*dimsys);
    equlast : natural; -- index of the last meaningful equation
    fail : boolean := false;

  begin
    put(file,"Number of equations : "); put(file,nbequ,1); new_line(file);
    put(file,"Number of unknowns  : "); put(file,nbunk,1); new_line(file);
    put(file,"I am assuming the top dimension is ");
    put(file,dimsys,1); put_line(file,"...");
    Add_Embed_Symbols(dimsys);
    wrk(p'range) := p;
    equlast := p'last;
    for i in 1..dimsys-1 loop
      Eliminate
        (file,wrk(1..equlast),dimj0,nbunk,dimsys,i,tol,constraint,fail);
      if fail
       then put_line(file,"Elimination failed, sorry.  Try again later...");
            exit;
       elsif i > 1
           then newcons := Add_Variables(constraint,i-1);
                Copy(newcons,constraint); Clear(newcons);
      end if;
      Differentiate(file,constraint,dimj0,nbunk,i,diffcons);
      if i > 1
       then equlast := equlast + 1;
            wrk(equlast) := Add_Variables(constraint,nbunk-dimj0);
            Clear(constraint);
      end if;
      equlast := equlast+1;
      Copy(diffcons,wrk(equlast));
      Clear(diffcons);
    end loop;
  end Diff_Elim;

  procedure Main is

    file : file_type;
    lp : Standard_Complex_Poly_Systems.Link_to_Poly_Sys;
    dimj0 : natural;

  begin
    new_line;
    put_line("Numerical Completion of a Differential Algebraic Equation.");
    new_line;
    get(lp);
    new_line;
    put_line("Reading the name of the output file.");
    Read_Name_and_Create_File(file);
    put(file,lp.all);
    new_line;
    Determine_Order(lp.all);
    new_line(file);
    put(file,"The order of symbols : ");
    Write_Symbol_Table(file);
    new_line(file);
    put("Give the number of original variables in J^0 : "); get(dimj0);
    put("Assuming symbol of derivative of variable i is at place i+");
    put(dimj0,1); put_line("...");
    new_line;
    put_line("See the output file for results...");
    new_line;
    Diff_Elim(file,lp.all,dimj0);
  end Main;

begin
  Main;
end ts_dae;
