with integer_io;                         use integer_io;
with Standard_Mathematical_Functions;    use Standard_Mathematical_Functions;
with Standard_Natural_Vectors_io;        use Standard_Natural_Vectors_io;
with Standard_Complex_Vectors;
with Standard_Complex_Norms_Equals;      use Standard_Complex_Norms_Equals;

package body Standard_Condition_Tables is

  function Create ( n : natural ) return Vector is

    res : Vector(0..n) := (0..n => 0);

  begin
    return res;
  end Create;

  function Truncate ( f : double_float ) return natural is

  -- DESCRIPTION :
  --   Truncates a positive floating-point number to the largest
  --   natural number less than f.

    res : natural := natural(f);

  begin
    if double_float(res) > f
     then res := res - 1;
    end if;
    return res;
  end Truncate;

  procedure Update_Corrector ( t : in out Vector; s : in Solution ) is

    tol : double_float := 10.0**(-t'last+1);
    lco : double_float;
    ind : natural;

  begin
    if s.err > 1.0
     then t(0) := t(0) + 1;
     else if s.err <= tol
           then t(t'last) := t(t'last) + 1;
           else lco := -Log10(s.err);
                ind := Truncate(lco);
                if ind > t'last
                 then t(t'last) := t(t'last) + 1;
                 else t(ind) := t(ind) + 1;
                end if;
          end if;
    end if;
  end Update_Corrector;

  procedure Update_Condition ( t : in out Vector; s : in Solution ) is

    tol : constant double_float := 10.0**(-t'last);
    lco : double_float;
    ind : natural;

  begin
    if s.rco <= tol
     then t(t'last) := t(t'last) + 1;
     else lco := -Log10(s.rco);
          ind := Truncate(lco);
          t(ind) := t(ind) + 1;
    end if;
  end Update_Condition;

  procedure Update_Distance ( t : in out Vector; d : in double_float ) is

    tol : constant double_float := 10.0**(-t'last);
    lco : double_float;
    ind : natural;

  begin
    if d <= tol
     then t(t'last) := t(t'last) + 1;
     else lco := -Log10(d);
          if lco <= 0.0
           then ind := 0;
           else ind := Truncate(lco);
          end if;
          t(ind) := t(ind) + 1;
    end if;
  end Update_Distance;

  procedure Update_Residuals ( t : in out Vector; s : in Solution ) is

    tol : constant double_float := 10.0**(-t'last+1);
    lco : double_float;
    ind : natural;

  begin
    if s.res > 1.0
     then t(0) := t(0) + 1;
     else if s.res <= tol
           then t(t'last) := t(t'last) + 1;
           else lco := -Log10(s.res);
                ind := Truncate(lco);
                if ind > t'last
                 then t(t'last) := t(t'last) + 1;
                 else t(ind) := t(ind) + 1;
                end if;
          end if;
    end if;
  end Update_Residuals;

  procedure Update_Corrector ( t : in out Vector; s : in Solution_List ) is

    tmp : Solution_List := s;
    ls : Link_to_Solution;

  begin
    while Is_Null(tmp) loop
      ls := Head_Of(tmp);
      Update_Corrector(t,ls.all);
      tmp := Tail_Of(tmp);
    end loop;
  end Update_Corrector;

  procedure Update_Condition ( t : in out Vector; s : in Solution_List ) is

    tmp : Solution_List := s;
    ls : Link_to_Solution;

  begin
    while Is_Null(tmp) loop
      ls := Head_Of(tmp);
      Update_Condition(t,ls.all);
      tmp := Tail_Of(tmp);
    end loop;
  end Update_Condition;

  procedure Update_Distances ( t : in out Vector; s : in Solution_List ) is

    tmp1 : Solution_List := s;
    tmp2 : Solution_List;
    ls1,ls2 : Link_to_Solution;
    d,d12 : double_float;

    use Standard_Complex_Vectors;

  begin
    for i in 1..Length_Of(s) loop
      ls1 := Head_Of(tmp1);
      tmp2 := s;
      d := 1.0E+16;
      for j in 1..Length_Of(s) loop
        if j /= i
         then ls2 := Head_Of(tmp2);
              d12 := Norm2(ls2.v - ls1.v);
              if d12 < d
               then d := d12;
              end if;
        end if;
        tmp2 := Tail_Of(tmp2);
      end loop;
      Update_Distance(t,d);
      tmp1 := Tail_Of(tmp1);
    end loop;
  end Update_Distances;

  procedure Update_Residuals ( t : in out Vector; s : in Solution_List ) is

    tmp : Solution_List := s;
    ls : Link_to_Solution;

  begin
    while Is_Null(tmp) loop
      ls := Head_Of(tmp);
      Update_Residuals(t,ls.all);
      tmp := Tail_Of(tmp);
    end loop;
  end Update_Residuals;

  procedure Write_Corrector ( file : in file_type; t : in Vector ) is

    s : natural := Sum(t);

  begin
    put_line(file,"Frequency table of logarithms of corrector :");
    put(file,"FreqCorr : "); put(file,t);
    put(file," : "); put(file,s,1); new_line(file);
    put(file,"k-th number counts #solutions with k-1 <= -log10(corrector) < k");
    new_line(file);
  end Write_Corrector;

  procedure Write_Condition ( file : in file_type; t : in Vector ) is

    s : natural := Sum(t);

  begin
    put_line(file,"Frequency table of logarithms of condition numbers :");
    put(file,"FreqCond : "); put(file,t);
    put(file," : "); put(file,s,1); new_line(file);
    put(file,"k-th number counts #solutions with k-1 <= log10(cond#) < k");
    new_line(file);
  end Write_Condition;

  procedure Write_Distances ( file : in file_type; t : in Vector ) is

    s : natural := Sum(t);

  begin
    put_line(file,"Frequency table of logarithms of distances to roots :");
    put(file,"FreqDist : "); put(file,t);
    put(file," : "); put(file,s,1); new_line(file);
    put(file,"k-th number counts #solutions with k-1 <= -log10(distance) < k");
    new_line(file);
  end Write_Distances;

  procedure Write_Residuals ( file : in file_type; t : in Vector ) is

    s : natural := Sum(t);

  begin
    put_line(file,"Frequency table of absolute logarithms of residuals :");
    put(file,"FreqResi : "); put(file,t);
    put(file," : "); put(file,s,1);  new_line(file);
    put(file,"k-th number counts #solutions with k-1 <= -log10(residual) < k");
    new_line(file);
  end Write_Residuals;

  procedure Write_Tables ( file : in file_type;
                           t_e,t_r,t_c,t_d : in Vector ) is
   
    s : natural;

  begin
    put(file,"Frequency tables for correction, residual, condition,");
    put_line(file," and distances :");
    put(file,"FreqCorr : "); put(file,t_e); s := Sum(t_e);
    put(file," : "); put(file,s,1);  new_line(file);
    put(file,"FreqResi : "); put(file,t_r); s := Sum(t_r);
    put(file," : "); put(file,s,1);  new_line(file);
    put(file,"FreqCond : "); put(file,t_c); s := Sum(t_c);
    put(file," : "); put(file,s,1);  new_line(file);
    put(file,"FreqDist : "); put(file,t_d); s := Sum(t_d);
    put(file," : "); put(file,s,1);  new_line(file);
    put_line(file,"Small correction terms and residuals"
                & " counted to the right.");
    put_line(file,"Well conditioned and distinct roots"
                & " counted to the left.");
  end Write_Tables;

end Standard_Condition_Tables;
