with Ada.Calendar;                       use Ada.Calendar;
with text_io,integer_io;                 use text_io,integer_io;
with Communications_with_User;           use Communications_with_User;
with File_Scanning;                      use File_Scanning;
with Timing_Package,Time_Stamps;         use Timing_Package,Time_Stamps;
with Standard_Complex_Polynomials;       use Standard_Complex_Polynomials;
with Standard_Complex_Poly_Systems;      use Standard_Complex_Poly_Systems;
with Standard_Complex_Poly_Systems_io;   use Standard_Complex_Poly_Systems_io;
with Standard_Complex_Solutions;         use Standard_Complex_Solutions;
with Standard_Complex_Solutions_io;      use Standard_Complex_Solutions_io;
with Standard_Scaling;                   use Standard_Scaling;
with Black_Box_Durand_Kerner;
with Black_Box_Fewnomial_Solver;
with Black_Box_Factorization;
with Black_Box_Root_Counting;
with Black_Box_Poly_Continuations;       use Black_Box_Poly_Continuations;
with bablsolve;

procedure bablphc ( infilename,outfilename : in string ) is

  start_moment : Time := Clock;
  ended_moment : Time;

  procedure Read_System ( file : in out file_type; filename : in string;
                          p : out Link_to_Poly_Sys ) is
  begin
    if filename /= ""
     then Open_Input_File(file,filename);
          get(file,p);
    end if;
  exception
    when others => put_line("Something is wrong with argument file...");
                   p := null; return;
  end Read_System;

  procedure Timing_Summary ( file : in file_type;
                             roco,hoco,poco,total : in duration ) is

    b0 : constant string :=
     "  ---------------------------------------------------------------------";
    b1 : constant string :=
     "  |                    TIMING INFORMATION SUMMARY                     |";
    b2 : constant string :=
     "  |   root counts  |  start system  |  continuation  |   total time   |";

  begin
    put_line(file,b0);
    put_line(file,b1);
    put_line(file,b0);
    put_line(file,b2);
    put_line(file,b0);
    put(file,"  | ");
    print_hms(file,roco); put(file," | ");
    print_hms(file,hoco); put(file," | ");
    print_hms(file,poco); put(file," | ");
    print_hms(file,total); put_line(file," |");
    put_line(file,b0);
  end Timing_Summary;

  procedure Single_Main ( p : in Poly ) is

  -- DESCRIPTION :
  --   This procedure deals with one single polynomial.

    n : constant natural := Number_of_Unknowns(p);
    outfile : file_type;

  begin
   -- put("we have one single polynomial in ");
   -- put(n,1); put_line(" variable(s)...");
    Create_Output_File(outfile,outfilename);
    if n = 1
     then Black_Box_Durand_Kerner(outfile,p);
     else Black_Box_Factorization(infilename,outfile,p);
    end if;
  end Single_Main;

  procedure Square_Main
              ( infile : in out file_type; p : in out Link_to_Poly_Sys;
                append_sols : in boolean ) is

  -- DESCRIPTION :
  --   This procedure deals with general polynomial systems,
  --   should only be called for square polynomial systems.

    timer : timing_widget;
    outfile : file_type;
    q : Link_to_Poly_Sys;
    rc : natural;
    sols : Solution_List;
    roco,hoco,poco,total : duration := 0.0;
    fail : boolean;

  begin
    Create_Output_File(outfile,outfilename);
    put(outfile,p'last,p.all);
    tstart(timer);
    Black_Box_Fewnomial_Solver(outfile,p.all,sols,fail);
    if fail
     then
       q := new Poly_Sys(p'range);
       declare
         pp : Poly_Sys(p'range) := p.all;
       begin
         Black_Box_Root_Counting(outfile,pp,false,rc,q.all,sols,roco,hoco);
         if rc /= 0
          then Scale(pp);
               Black_Box_Polynomial_Continuation(outfile,pp,q.all,sols,poco);
         end if;
       end;
    end if;
    tstop(timer);
    total := Elapsed_User_Time(timer);
    new_line(outfile);
    print_times(outfile,timer,"Solving the polynomial system");
    new_line(outfile);
    Timing_Summary(outfile,roco,hoco,poco,total);
    ended_moment := Clock;
    put(outfile,"PHC ran from "); Write_Time_Stamp(outfile,start_moment);
    put(outfile," till ");  Write_Time_Stamp(outfile,ended_moment);
    put_line(outfile,".");
    Write_Elapsed_Time(outfile,start_moment,ended_moment);
    Close(outfile);
    if not Is_Null(sols) and append_sols
     then Close(infile);
          Open_Append_File(infile,infilename);
          new_line(infile);
          put_line(infile,"THE SOLUTIONS :");
          put(infile,Length_Of(sols),Head_Of(sols).n,sols);
          Close(infile);
    end if;
  end Square_Main;

  procedure Main is

    infile : file_type;
    append_sols : boolean := false;
    p : Link_to_Poly_Sys;
    n : natural;

  begin
    Read_System(infile,infilename,p);
    if p = null
     then new_line; get(p);
     else Scan_and_Skip(infile,"SOLUTIONS",append_sols);
          append_sols := not append_sols;
    end if;
    n := Number_of_Unknowns(p(p'first));
    if p'last = p'first then
      Single_Main(p(p'first));
    elsif p'last = n then
      Square_Main(infile,p,append_sols);
    else
      bablsolve(p.all);
    end if;
  end Main;

begin
  Main;
end bablphc;
