with text_io;                            use text_io;
with Unix_Command_Line;
with Standard_Random_Numbers; -- to handle zero seed

with mainscal,mainred;        -- scaling and reduction
with mainroco,bablroco;       -- general root counting
with mainsmvc,babldmvc;       -- mixed-volume computation
with mainpoco,bablpoco;       -- polynomial continuation
with mainphc,bablphc;         -- main phc driver + blackbox
with mainvali,bablvali;       -- validation tool
with mainenum,bablenum;       -- numerical Schubert calculus
with mainfeed;                -- realization of pole placing feedback
with mainzip;                 -- putting solutions in Maple format
with mainfac,maindeco;        -- factorization and decomposition
with mainsolve;               -- equation-by-equation solver
with mainwit;                 -- witness set intersection
with mainhyp;                 -- witness set for hypersurface
with Split_or_Join_Solution_Lists;

-- NOTICE: (added for pieri_solver.ali)

with Interfaces.C;
with Complex_Polynomial_Matrices;        use Complex_Polynomial_Matrices;
with Complex_Polynomial_Matrices_io;     use Complex_Polynomial_Matrices_io;
with Verify_Solution_Maps;
with C_to_Ada_Arrays;                    use C_to_Ada_Arrays;

procedure Dispatch is

-- BANNERS WITH INFORMATION TO START DIALOGUE WITH USER :

  welcome : constant string :=
"Welcome to PHC (Polynomial Homotopy Continuation) V2.3.01 15 July 2005";

  author : constant string :=
    "Author is Jan Verschelde (E-mail: na.jverschelde@na-net.ornl.gov).";

  scalban : constant string :=
    "Equation/variable Scaling on polynomial system and solution list.";

  reduban : constant string :=
    "Linear and nonlinear Reduction w.r.t the total degree of the system.";

  rocoban : constant string :=
    "Root counting and Construction of product and polyhedral start systems.";

  mvcban : constant string :=
    "Mixed-Volume Computation by four different lifting strategies.";

  pocoban : constant string :=
    "Polynomial Continuation defined by a homotopy in one parameter.";

  valiban : constant string :=
    "Validation, refinement and purification of computed solution lists.";

  enumban : constant string :=
    "SAGBI/Pieri homotopies to solve a problem in enumerative geometry.";

  compban : constant string :=
    "A numerical irreducible decomposition for the solution components.";

  facban : constant string :=
    "Factor a pure dimensional solution set into irreducible components.";

  witban : constant string :=
    "Witness Set Intersection using Diagonal Homotopies";

  hypban : constant string :=
    "Witness Set for Hypersurface cutting with Random Line";

-- AVAILABLE OPTIONS :

  options : constant string := "0asdpmrvbekczftwl";
  -- 0 : zero seed for repeatable runs
  -- a : solve => equation-by-equation solver
  -- b : batch or black box processing
  -- c : comp => numerical irreducible decomposition
  -- d : redu => reduction w.r.t. the total degree
  -- e : enum => numerical Schubert calculus
  -- f : fac  => factor pure dimensional solution set into irreducibles
  -- k : feba => dynamic output feedback to control linear systems
  -- l : hyp  => witness set for hypersurface cutting with random line
  -- m : mvc  => mixed-volume computation
  -- p : poco => polynomial continuation
  -- r : roco => root counting methods
  -- s : scal => scaling of a polynomial system
  -- t : tool => tool to split or join solution lists
  -- v : vali => validation of solutions
  -- w : wit  => intersection of witness sets using diagonal homotopies
  -- z : zip  => strip output for solutions into Maple format

  argc : natural := Unix_Command_Line.Number_of_Arguments;

-- UTILITIES FOR PROCESSING THE ARGUMENTS AND OPTIONS :

  function Read_Argument ( k : in natural ) return string is

  -- DESCRIPTION :
  --   Reads the kth argument from the command line.
  --   An argument is a string not proceeded by a `-' character.
  --   The empty string is returned when there is no argument.

    null_string : constant string := "";
    cnt : natural := 0;

  begin
    if argc >= 1
     then for i in 1..argc loop
            declare
              s : constant string := Unix_Command_Line.Argument(i);
            begin
              if s(1) /= '-'
               then cnt := cnt + 1;
                    if k = cnt
                     then return s;
                    end if;
              end if;
            end;
          end loop;
    end if;
    return null_string;
  end Read_Argument;

  function Position ( c : character; s : string ) return natural is

  -- DESCRIPTION :
  --   If the the string contains the character c, then its position
  --   in the string will be returned.  Otherwise s'first-1 will be returned.

  begin
    for i in s'range loop
      if s(i) = c
       then return i;
      end if;
    end loop;
    return s'first-1;
  end Position;

  procedure Read_Next_Option ( pos : in out natural; legal : in string;
                               option : out character ) is

  -- DESCRIPTION :
  --   Reads the next option from the command line arguments.

  -- ON ENTRY :
  --   pos      position in the command line of the last option
  --            that has been read;
  --   legal    string which contains all legal options.

  -- ON RETURN :
  --   pos      the position in the command line of the last option read;
  --   option   is blank when no legal option could be read, otherwise it
  --            contains the next legal option.

    res : character := ' ';
    start : natural := pos+1;

  begin
    if argc >= 1
     then for i in start..argc loop
            declare
              s : constant string := Unix_Command_Line.Argument(i);
            begin
              if s(1) = '-'
               then pos := Position(s(2),legal);
                    if pos >= legal'first
                     then res := legal(pos);
                     else put("The option '"); put(s);
                          put_line("' is not recognised.  Will ignore it...");
                    end if;
              end if;
            end;
            pos := i;
            exit when (res /= ' ');
          end loop;
    end if;
    option := res;
  end Read_Next_Option;

-- DISPATCHING ACCORDING TO OPTIONS :

  procedure Black_Box_Dispatcher
              ( option2 : in character;
                file1,file2,file3 : in string ) is

  -- DESCRIPTION :
  --   When the first option is 'b', then this routine handles the
  --   second option and calls the appropriate main driver.

  begin
    case option2 is
      when 's'    => mainscal(file1,file2);
      when 'd'    => mainred(file1,file2);
      when 'r'    => bablroco(file1,file2);
      when 'm'    => babldmvc(file1,file2);
      when 'p'    => bablpoco(file1,file2,file3);
      when 'v'    => bablvali(file1,file2);
      when 'e'    => bablenum(file1,file2);
      when others => bablphc(file1,file2);
    end case;
  end Black_Box_Dispatcher;

  procedure Scaling_Dispatcher ( infile,outfile : in string ) is

  -- DESCRIPTION :
  --   This routine is called when the first option is 's'.

  begin
    put_line(welcome); put_line(scalban);
    mainscal(infile,outfile);
  end Scaling_Dispatcher;

  procedure Reduction_Dispatcher ( infile,outfile : in string ) is

  -- DESCRIPTION :
  --   This dispatcher is called when the first option is 'd'.

  begin
    put_line(welcome); put_line(reduban);
    mainred(infile,outfile);
  end Reduction_Dispatcher;

  procedure Root_Count_Dispatcher
              ( o2 : in character; infile,outfile : in string ) is

  -- DESCRIPTION :
  --   Calls the root couting facilities in PHCpack,
  --   when the first option was 'r'.

  begin
    case o2 is
      when 'b'    => bablroco(infile,outfile);
      when others => put_line(welcome); put_line(rocoban);
                     mainroco(infile,outfile);
    end case;
  end Root_Count_Dispatcher;

  procedure Mixed_Volume_Dispatcher
              ( o2 : in character; infile,outfile : in string ) is

  -- DESCRIPTION :
  --   Calls the mixed-volume computation menu.

  begin
    case o2 is
      when 'b'    => babldmvc(infile,outfile);
      when others => put_line(welcome); put_line(mvcban);
                     mainsmvc(infile,outfile);
    end case;
  end Mixed_Volume_Dispatcher;

  procedure Continuation_Dispatcher 
              ( o2 : in character; file1,file2,file3 : in string ) is

  -- DESCRIPTION :
  --   Invokes the path trackers in PHCpack.

  begin
    case o2 is
      when 'b'    => bablpoco(file1,file2,file3);
      when others => put_line(welcome); put_line(pocoban);
                     mainpoco(file1,file2);
    end case;
  end Continuation_Dispatcher;

  procedure Validation_Dispatcher
              ( o2 : in character; infile,outfile : in string ) is

  -- DESCRIPTION :
  --   Validates the computed results.

  begin
    case o2 is
      when 'b'    => bablvali(infile,outfile);
      when others => put_line(welcome); put_line(valiban);
                     mainvali(infile,outfile);
    end case;
  end Validation_Dispatcher;

  procedure Enumeration_Dispatcher
              ( o2 : in character; infile,outfile : in string ) is

  -- DESCRIPTION :
  --   Invokes the numerical Schubert calculus.

  begin
    case o2 is
      when 'b'    => bablenum(infile,outfile);
      when others => put_line(welcome); put_line(enumban);
                     mainenum;
    end case;
  end Enumeration_Dispatcher;

  procedure Decomposition_Dispatcher ( infile,outfile : in string ) is

  -- DESCRIPTION :
  --   Witness point generation facilities.

  begin
    put_line(welcome); put_line(compban);
    maindeco(infile,outfile);
  end Decomposition_Dispatcher;

  procedure Factorization_Dispatcher ( infile,outfile : in string ) is

  -- DESCRIPTION :
  --   Filtering and factorization capabilities.

  begin
    put_line(welcome); put_line(facban);
    mainfac(infile,outfile);
  end Factorization_Dispatcher;

  procedure Witness_Set_Intersection_Dispatcher
              ( witset_one,witset_two,logfile : in string ) is

  -- DESCRIPTION :
  --   Intersection of two witness sets using diagonal homotopies.

  begin
    if witset_one = "" or witset_two = "" or logfile = ""
     then put_line(welcome); put_line(witban);
    end if;
    mainwit(witset_one,witset_two,logfile);
  end Witness_Set_Intersection_Dispatcher;

  procedure Witness_Set_for_Hypersurface_Dispatcher
              ( polyfile,logfile : in string ) is

  -- DESCRIPTION :
  --   Creation of witness set for hypersurface.

  begin
    if polyfile = "" or logfile = ""
     then put_line(welcome); put_line(hypban);
    end if;
    mainhyp(polyfile,logfile);
  end Witness_Set_for_Hypersurface_Dispatcher;

  procedure General_Dispatcher
              ( o1,o2 : in character; f1,f2,f3 : in string ) is

  -- DESCRIPTION :
  --   This is the general dispatcher for the first option o1.

  begin
    case o1 is
      when 'a'    => mainsolve(f1,f2);
      when 'b'    => Black_Box_Dispatcher(o2,f1,f2,f3);
      when 'c'    => Decomposition_Dispatcher(f1,f2);
      when 'd'    => Reduction_Dispatcher(f1,f2);
      when 'e'    => Enumeration_Dispatcher(o2,f1,f2);
      when 'm'    => Mixed_Volume_Dispatcher(o2,f1,f2);
      when 'k'    => mainfeed(f1,f2);
      when 'l'    => Witness_Set_for_Hypersurface_Dispatcher(f1,f2);
      when 's'    => Scaling_Dispatcher(f1,f2);
      when 'f'    => Factorization_Dispatcher(f1,f2);
      when 'p'    => Continuation_Dispatcher(o2,f1,f2,f3);
      when 'v'    => Validation_Dispatcher(o2,f1,f2);
      when 'r'    => Root_Count_Dispatcher(o2,f1,f2);
      when 't'    => Split_or_Join_Solution_Lists;
      when 'w'    => Witness_Set_Intersection_Dispatcher(f1,f2,f3);
      when 'z'    => mainzip(f1,f2);
      when others => put_line(welcome); mainphc(f1,f2);
    end case;
  end General_Dispatcher;

  procedure Main is

  -- DESCRIPTION :
  --   Parses the arguments on the command line, separating the arguments
  --   into options (at most 2) and file names (at most 3).

    posi : natural := 0;
    option1,option2,option3 : character;
    file1 : constant string := Read_Argument(1);
    file2 : constant string := Read_Argument(2);
    file3 : constant string := Read_Argument(3);

  begin
    Read_Next_Option(posi,options,option1);
    if option1 /= ' '
     then Read_Next_Option(posi,options,option2);
          if option2 /= ' '
           then Read_Next_Option(posi,options,option3);        
           else option3 := ' ';
          end if;
     else option2 := ' ';
          option3 := ' ';
    end if;
    if (file1 /= "") and then (file1 = file2)
     then new_line;
          put_line("The first two file names are the same.");
          put_line("Will ignore the second file name...");
          if option1 = '0' then
            Standard_Random_Numbers.Set_Seed(0);
            General_Dispatcher(option2,option3,file1,"",file3);
          elsif option2 = '0' then
            Standard_Random_Numbers.Set_Seed(0);
            General_Dispatcher(option1,option3,file1,"",file3);
          elsif option3 = '0' then
            Standard_Random_Numbers.Set_Seed(0);
            General_Dispatcher(option1,option2,file1,"",file3);
          else
            General_Dispatcher(option1,option2,file1,"",file3);
          end if;
     else if option1 = '0' then
            Standard_Random_Numbers.Set_Seed(0);
            General_Dispatcher(option2,option3,file1,file2,file3);
          elsif option2 = '0' then
            Standard_Random_Numbers.Set_Seed(0);
            General_Dispatcher(option1,option3,file1,file2,file3);
          elsif option3 = '0' then
            Standard_Random_Numbers.Set_Seed(0);
            General_Dispatcher(option1,option2,file1,file2,file3);
          else
            General_Dispatcher(option1,option2,file1,file2,file3);
          end if;
    end if;
  end Main;

begin
  Main;
end Dispatch;
