with Ada.Calendar;                       use Ada.Calendar;
with text_io,integer_io;                 use text_io,integer_io;
with Communications_with_User;           use Communications_with_User;
with Timing_Package;                     use Timing_Package;
with Time_Stamps;                        use Time_Stamps;
with Standard_Complex_Numbers;           use Standard_Complex_Numbers;
with Standard_Complex_Vectors;           use Standard_Complex_Vectors;
with Standard_Complex_Polynomials;       use Standard_Complex_Polynomials;
with Standard_Complex_Poly_Systems;      use Standard_Complex_Poly_Systems;
with Standard_Complex_Poly_Systems_io;   use Standard_Complex_Poly_Systems_io;
with Standard_Complex_Solutions;         use Standard_Complex_Solutions;
with Multprec_Floating_Numbers;
with Multprec_Complex_Solutions;
with Drivers_for_Scaling;                use Drivers_for_Scaling;
with Drivers_for_Reduction;              use Drivers_for_Reduction;
with Drivers_for_Poly_Continuation;      use Drivers_for_Poly_Continuation;
with Driver_for_Root_Counts;
with Drivers_for_Homotopy_Creation;      use Drivers_for_Homotopy_Creation;
with Driver_for_Root_Refining;
with Bye_Bye_Message;

procedure mainphc ( infilename,outfilename : in string ) is

  start_moment : Time := Clock;
  ended_moment : Time;
  outpt : file_type;
  lp : Link_to_Poly_Sys := null;
  timer : timing_widget;

  procedure Display_Options is

  -- DESCRIPTION :
  --   Displays an overview of all options on screen.

    o : array(0..16) of string(1..65);

  begin
    put_line("Running in full mode.  Note also the following options:");
    o(0) := "  phc -0 : random numbers with zero seed for repeatable runs     ";
    o(1) := "  phc -a : Solving polynomial systems equation-by-equation       ";
    o(2) := "  phc -b : Batch or black-box processing                         ";
    o(3) := "  phc -c : Irreducible decomposition for solution components     ";
    o(4) := "  phc -d : Linear and nonlinear Reduction w.r.t. the total degree";
    o(5) := "  phc -e : SAGBI/Pieri homotopies to intersect linear subspaces  ";
    o(6) := "  phc -f : Factor pure dimensional solution set into irreducibles";
    o(7) := "  phc -k : realization of dynamic output feedback placing poles  ";
    o(8) := "  phc -l : Witness Set for Hypersurface cutting with Random Line ";
    o(9) := "  phc -m : Mixed-Volume Computation by four lifting strategies   ";
    o(10):= "  phc -p : Polynomial Continuation by a homotopy in one parameter";
    o(11):= "  phc -r : Root counting and Construction of start systems       ";
    o(12):= "  phc -s : Equation and variable Scaling on system and solutions ";
    o(13):= "  phc -t : tool to split or join solution lists                  ";
    o(14):= "  phc -v : Validation, refinement and purification of solutions  ";
    o(15):= "  phc -w : Witness Set Intersection using Diagonal Homotopies    ";
    o(16):= "  phc -z : strip phc output solution lists into Maple format     ";
    for i in o'range loop
      put_line(o(i));
    end loop;
    put_line("Options may be combined, e.g.: phc -b -0 or phc -0 -b.");
  end Display_Options;

  procedure Read_System ( filename : in string ) is

    file : file_type;

  begin
    if filename /= ""
     then Open_Input_File(file,filename);
          get(file,lp);
          Close(file);
    end if;
  exception
    when others => new_line;
                   put("Could not open file with name ");
                   put(filename); put(".");
                   lp := null; return;
  end Read_System;

begin
  new_line;
  Display_Options;
  Read_System(infilename);
  if lp = null
   then new_line; get(lp);
  end if;
  Create_Output_File(outpt,outfilename);
  put(outpt,lp'last,lp.all);
  tstart(timer);
  declare
    p,q,scalp,projp : Poly_Sys(lp'range);
    target : Complex_Number;
    basis,roco,deci,size : natural;
    scalvec : Link_to_Vector;
    sols : Solution_List;
    mpsols : Multprec_Complex_Solutions.Solution_List;
    proj : boolean;
  begin
    p := lp.all;  Copy(p,scalp);
    Driver_for_Scaling(outpt,scalp,basis,scalvec);
    Driver_for_Reduction(outpt,scalp,roco,true);
    Copy(scalp,projp);
    Driver_for_Root_Counts(outpt,projp,q,true,sols,roco);
    if Length_Of(sols) > 0
     then Driver_for_Homotopy_Construction(outpt,projp,q,sols,target,deci);
          proj := (Number_of_Unknowns(p(p'first)) > p'last);
          if Head_Of(sols).t /= Create(0.0)
           then Set_Continuation_Parameter(sols,Create(0.0));
          end if;
          if deci <= 16
           then Driver_for_Polynomial_Continuation(outpt,sols,proj,target);
                Driver_for_Root_Refining(outpt,scalp,p,basis,scalvec,sols);
           else mpsols := Multprec_Complex_Solutions.Create(sols);
                size := Multprec_Floating_Numbers.Decimal_to_Size(deci);
                Multprec_Complex_Solutions.Set_Size(mpsols,size);
                Driver_for_Polynomial_Continuation
                  (outpt,mpsols,proj,deci,target);
          end if;
     end if;
  end;
  tstop(timer);
  new_line(outpt);
  print_times(outpt,timer,"solving the polynomial system");
  new_line(outpt);
 -- put(outpt,Bye_Bye_Message);
  ended_moment := Clock;
  put(outpt,"PHC ran from "); Write_Time_Stamp(outpt,start_moment);
  put(outpt," till ");        Write_Time_Stamp(outpt,ended_moment);
  put_line(outpt,".");
  Write_Elapsed_Time(outpt,start_moment,ended_moment);
  Close(outpt);
end mainphc;
