with Standard_Common_Divisors;           use Standard_Common_Divisors;

package body Smith_Normal_Form is

-- NOTE :
--   The implementation consists in three parts:
--     1. The triangulation of one row and column by left and right
--        multiplication with unimodular matrices.
--     2. Finding pivots and permuting rows and columns.
--     3. Repeatedly invoking steps 1 and 2 yields the Smith normal form.

-- DYNAMIC TRIANGULATORS :

  procedure Upper_Triangulate ( u,a : in out Matrix; row,col : in integer ) is

  -- DESCRIPTION :
  --   Makes all elements under the current row in the indicated column zero,
  --   by application of unimodular transformations.

  -- REQUIRED : a(row,col) /= 0.

    aa,bb,ka,lb,d,a_rowk,a_ik : integer;

  begin
    for i in (row+1)..a'last(1) loop
      if a(i,col) /= 0                                -- make a(i,col) zero
       then gcd(a(row,col),a(i,col),ka,lb,d);         -- compute multipliers
            aa := a(row,col)/d;     bb := a(i,col)/d;
            if (aa = bb) and then (ka = 0)
             then ka := lb; lb := 0;
            end if;
            if (aa = -bb) and then (ka = 0)
             then ka := -lb; lb := 0;
            end if;
            for k in a'range(2) loop                  -- perform combinations
              a_rowk := a(row,k);
              a_ik := a(i,k);
              a(row,k) :=  ka*a_rowk + lb*a_ik;
              a(i,k)   := -bb*a_rowk + aa*a_ik;
            end loop;
            for k in u'range(2) loop
              a_rowk := u(row,k);
              a_ik := u(i,k);
              u(row,k) :=  ka*a_rowk + lb*a_ik;
              u(i,k)   := -bb*a_rowk + aa*a_ik;
            end loop;
      end if;
    end loop;
  end Upper_Triangulate;

  procedure Lower_Triangulate ( a,v : in out Matrix; row,col : in integer ) is

  -- DESCRIPTION :
  --   Makes all elements at the right of the current column in the indicated
  --   row zero, by application of unimodular transformations.

  -- REQUIRED : a(row,col) /= 0.

    aa,bb,ka,lb,d,a_kcol,a_kj : integer;

  begin
    for j in (col+1)..a'last(2) loop
      if a(row,j) /= 0                                 -- make a(row,j) zero
       then gcd(a(row,col),a(row,j),ka,lb,d);          -- compute multipliers
            aa := a(row,col)/d;        bb := a(row,j)/d;
            if (aa = bb) and then (ka = 0)
             then ka := lb; lb := 0;
            end if;
            if (aa = -bb) and then (ka = 0)
             then ka := -lb; lb := 0;
            end if;
            for k in a'range(1) loop                   -- perform combinations
              a_kcol := a(k,col);
              a_kj := a(k,j);
              a(k,col) := a_kcol*ka    + a_kj*lb;
              a(k,j)   := a_kcol*(-bb) + a_kj*aa;
            end loop;
            for k in v'range(1) loop
              a_kcol := v(k,col);
              a_kj := v(k,j);
              v(k,col) := a_kcol*ka    + a_kj*lb;
              v(k,j)   := a_kcol*(-bb) + a_kj*aa;
            end loop;
      end if;
    end loop;
  end Lower_Triangulate;

-- PIVOTING ROUTINES :

  function Absolute_Value ( i : in integer ) return integer is

  -- DESCRIPTION :
  --   Returns the absolute value of the integer.

  begin
    if i < 0
     then return -i;
     else return i;
    end if;
  end Absolute_Value;

  procedure Find_Pivots ( a : in Matrix; pivrow,pivcol : in out integer ) is

  -- DESCRIPTION :
  --   Finds the smallest nonzero entry in the matrix a, starting at
  --   the current pivrow and pivcol.

    row : integer := pivrow;
    col : integer := pivcol;
    min : integer := 20000000;

  begin
    for i in pivrow..a'last(1) loop
      for j in pivcol..a'last(2) loop
        if (a(i,j) /= 0) and then (Absolute_Value(a(i,j)) < min)
         then min := Absolute_Value(a(i,j));
              row := i; col := j;
        end if;
      end loop;
    end loop;
    pivrow := row;
    pivcol := col;
  end Find_Pivots;

  procedure Switch_Rows_and_Columns
              ( a,u,v : in out Matrix; pivrow,row,pivcol,col : in integer ) is

  -- DESCRIPTION :
  --   Switches the pivot rows and columns in a to the current row and column.
  --   The unimodular matrices u and v are pivoted along.

  -- ON ENTRY :
  --   a        matrix diagonal up to (row,col), (row,col) not included;
  --   u        unimodular matrix for left multiplication;
  --   v        unimodular matrix for right multiplication;
  --   pivrow   pivot row;
  --   row      current row;
  --   pivcol   pivot column;
  --   col      current column.

  -- ON RETURN :
  --   a        matrix with (pivrow,pivcol) at place of (row,col);
  --   u        updated unimodular matrix for left multiplication;
  --   v        updated unimodular matrix for right multiplication.

    temp : integer;

  begin
    if pivrow /= row                             -- interchange rows
     then for k in a'range(2) loop
            temp := a(row,k);
            a(row,k) := a(pivrow,k);
            a(pivrow,k) := temp;
          end loop;
          for k in u'range(2) loop
            temp := u(row,k);
            u(row,k) := u(pivrow,k);
            u(pivrow,k) := temp;
          end loop;
    end if;
    if pivcol /= col                            -- interchange columns
     then for k in a'range(1) loop
            temp := a(k,col);
            a(k,col) := a(k,pivcol);
            a(k,pivcol) := temp;
          end loop;
          for k in v'range(1) loop
            temp := v(k,col);
            v(k,col) := v(k,pivcol);
            v(k,pivcol) := temp;
          end loop;
    end if;
  end Switch_Rows_and_Columns;

-- TARGET ROUTINES :

  function Identity ( n : natural ) return Matrix is

    res : Matrix(1..n,1..n);

  begin
    for i in 1..n loop
      for j in 1..n loop
        if i = j
         then res(i,j) := 1;
         else res(i,j) := 0;
        end if;
      end loop;
    end loop;
    return res;
  end Identity;

  function Diagonal ( mat : Matrix ) return boolean is

  -- DESCRIPTION :
  --   Returns true if the matrix is diagonal, returns false otherwise.

  begin
    for i in mat'range(1) loop
      for j in mat'range(2) loop
        if i /= j
         then if mat(i,j) /= 0
               then return false;
              end if;
        end if;
      end loop;
    end loop;
    return true;
  end Diagonal;

  procedure Diagonalize ( u,a,v : in out Matrix ) is

  -- DESCRIPTION :
  --   Returns the Smith normal form of the matrix, d = u*a*v.

  -- ON ENTRY :
  --   u        identity matrix of range a'range(1);
  --   a        matrix that is to diagonalize;
  --   v        identity matrix of range a'range(2).

  -- ON RETURN :
  --   u        unimodular matrix of left multiplications;
  --   a        diagonal matrix, equals u*a*v;
  --   v        unimodular matrix of right multiplications.

    row : integer := a'first(1);
    col : integer := a'first(2);
    pivrow,pivcol : integer;

  begin
    while not Diagonal(a) loop
      row := a'first(1);
      col := a'first(2);
      loop
        pivrow := row; pivcol := col;
        Find_Pivots(a,pivrow,pivcol);
        exit when (a(pivrow,pivcol) = 0);
        Switch_Rows_and_Columns(a,u,v,pivrow,row,pivcol,col);
        Upper_Triangulate(u,a,row,col);
        Lower_Triangulate(a,v,row,col);
        row := row+1;
        col := col+1;
        exit when ((row > a'last(1)) or (col > a'last(2)));
      end loop;
    end loop;
  end Diagonalize;

end Smith_Normal_Form;
