with integer_io;                         use integer_io;
with Standard_Floating_Numbers;          use Standard_Floating_Numbers;
with Standard_Floating_Numbers_io;       use Standard_Floating_Numbers_io;
with Multprec_Floating_Numbers_io;       use Multprec_Floating_Numbers_io;
with Multprec_Complex_Numbers;           use Multprec_Complex_Numbers;
with Multprec_Complex_Numbers_io;        use Multprec_Complex_Numbers_io;
with Standard_Natural_Vectors;
with Standard_Floating_Vectors;
with Multprec_Complex_Vectors_io;        use Multprec_Complex_Vectors_io;
with Standard_Natural_Vectors;
with Multprec_Complex_Matrices;          use Multprec_Complex_Matrices;
with Multprec_Complex_Linear_Solvers;    use Multprec_Complex_Linear_Solvers;
with Multprec_Complex_Norms_Equals;      use Multprec_Complex_Norms_Equals;
with Multprec_Complex_Solutions_io;      use Multprec_Complex_Solutions_io;
with Multprec_Condition_Tables;          use Multprec_Condition_Tables;

package body Multprec_Root_Refiners is

-- AUXILIARIES :

  function Is_Real ( sol : Solution; tol : Floating_Number ) return boolean is

    res : boolean := true;

  begin
    for i in sol.v'range loop
      declare
        imapar : Floating_Number := IMAG_PART(sol.v(i));
        abstmp : Floating_Number := AbsVal(imapar);
      begin
        if abstmp > tol
         then res := false;
        end if;
        Clear(imapar); Clear(abstmp);
      end;
      exit when not res;
    end loop;
    return res;
  end Is_Real;

  function Equal ( s1,s2 : Solution; tol : Floating_Number ) return boolean is
  begin
    for i in s1.v'range loop
      if not Equal(s1.v(i),s2.v(i),tol)
       then return false;
      end if;
    end loop;
    return true;
  end Equal;

  function Is_Clustered ( sol : Solution; nb : natural; sols : Solution_List;
                          tol : Floating_Number ) return natural is

    tmp : Solution_List := sols;
    cnt : natural := 0;

  begin
    while not Is_Null(tmp) loop
      cnt := cnt + 1;
      if cnt /= nb then
        if Equal(sol,Head_Of(tmp).all,tol)
         then return cnt;
        end if;
      end if;
      tmp := Tail_Of(tmp);
    end loop;
    return nb;
  end Is_Clustered;

  function Is_Clustered ( sol : Solution; nb : natural; sols : Solution_Array;
                          tol : Floating_Number ) return natural is
  begin
    for i in sols'range loop
      if i /= nb then
        if Equal(sol,sols(i).all,tol)
         then return i;
        end if;
      end if;
    end loop;
    return nb;
  end Is_Clustered;

  function Multiplicity ( sol : Solution; sols : Solution_List; 
                          tol : Floating_Number ) return natural is

    tmp : Solution_List := sols;
    cnt : natural := 0;

  begin
    while not Is_Null(tmp) loop
      if Equal(sol,Head_Of(tmp).all,tol)
       then cnt := cnt + 1;
      end if;
      tmp := Tail_Of(tmp);
    end loop;
    return cnt;
  end Multiplicity;

  function Multiplicity ( sol : Solution; sols : Solution_Array;
                          tol : Floating_Number ) return natural is
    cnt : natural := 0;

  begin
    for i in sols'range loop
      if Equal(sol,sols(i).all,tol)
       then cnt := cnt + 1;
      end if;
    end loop;
    return cnt;
  end Multiplicity;

  procedure Write_Bar ( file : in file_type ) is
  begin
    for i in 1..75 loop
      put(file,'=');
    end loop;
    new_line(file);
  end Write_Bar;

  procedure Write_Info ( file : in file_type; zero : in Solution;
                         initres : double_float;
                         i,numb : in natural; fail : in boolean ) is

  -- DESCRIPTION :
  --   The information concerning the zero is written

  begin
    put(file,"solution "); put(file,i,1); put(file," : ");
    put(file,"   start residual : "); put(file,initres,2,3,3);
    put(file,"   #iterations : "); put(file,numb,1);
    if fail
     then put_line(file,"   failure");
     else put_line(file,"   success");
    end if;
    put(file,zero);
  end Write_Info;
 
  procedure Root_Accounting
               ( file : in file_type; ls : in out Link_to_Solution;
                 nb : in natural; sa : in out Solution_Array;
                 fail : in boolean; tolsing,tolclus : in Floating_Number;
                 nbfail,nbreal,nbcomp,nbreg,nbsing,nbclus : in out natural ) is

  -- DESCRIPTION :
  --   This procedure does root accounting of the solution sol, w.r.t. the
  --   solution list sols.  Information will be provided concerning the type
  --   of solution.

    tolreal : Floating_Number := Create(1.0E-13);

  begin
    if fail then
      put_line(file," no solution ==");
      nbfail := nbfail + 1;
      ls.m := 0;
    else 
      if Is_Real(ls.all,tolreal)
       then put(file," real ");
            nbreal := nbreal + 1;
       else put(file," complex ");
            nbcomp := nbcomp + 1;
      end if;
      if sa(nb).rco < tolsing then
        declare
          m : natural := Multiplicity(ls.all,sa,tolclus);
        begin
          if m = 1
           then m := 0;
          end if;
          ls.m := m;
        end;
        put_line(file,"singular ==");
        nbsing := nbsing + 1;
      else
        declare
          nb2 : natural := Is_Clustered(ls.all,nb,sa,tolclus);
        begin
          if nb2 = nb
           then put_line(file,"regular ==");
                nbreg := nbreg + 1;
               -- ls.m := 1;
           else put(file,"clustered : ");
                put(file,nb2,1);
                put_line(file," ==");
                nbclus := nbclus + 1;
                ls.m := -nb2;
                sa(nb2).m := -nb;
          end if;
        end;	   
      end if;  
    end if;
    Clear(tolreal);
  end Root_Accounting;

  procedure Root_Accounting 
                ( ls : in out Link_to_Solution; nb : in natural;
                  sa : in out Solution_Array; fail : in boolean;
                  tolsing,tolclus : in Floating_Number ) is

  -- DESCRIPTION :
  --   This procedure does root accounting of the solution sol, w.r.t. the
  --   solution list sols.  Information will be provided concerning the type
  --   of solution.

  begin
    if fail then
      ls.m := 0;
    elsif sa(nb).rco < tolsing then
      declare
        m : natural := Multiplicity(ls.all,sa,tolclus);
      begin
        if m = 1
         then ls.m := 0;
         else ls.m := m;
        end if;
      end;
    else -- sa(nb).rco > tolsing, check for clustering
      declare
        nb2 : natural := Is_Clustered(ls.all,nb,sa,tolclus);
      begin
        if nb2 /= nb
         then ls.m := -nb2;
              sa(nb2).m := -nb;
        end if;
       -- if nb2 = nb
       --  then ls.m := 1;
       --  else ls.m := -nb2;
       -- end if;
      end;	   
    end if;
  end Root_Accounting;

  procedure Write_Global_Info
             ( file : in file_type;
               tot,nbfail,nbreal,nbcomp,nbreg,nbsing,nbclus : in natural ) is

  begin
    Write_Bar(file);
    put(file,"A list of "); put(file,tot,1);
    put_line(file," solutions has been refined :");
    put(file,"Number of regular solutions   : "); put(file,nbreg,1);
    put_line(file,".");
    put(file,"Number of singular solutions  : "); put(file,nbsing,1);
    put_line(file,".");
    put(file,"Number of real solutions      : "); put(file,nbreal,1);
    put_line(file,".");
    put(file,"Number of complex solutions   : "); put(file,nbcomp,1);
    put_line(file,".");
    put(file,"Number of clustered solutions : "); put(file,nbclus,1);
    put_line(file,".");
    put(file,"Number of failures            : "); put(file,nbfail,1);
    put_line(file,".");
    Write_Bar(file);
  end Write_Global_Info;

-- ESTIMATOR OF MULTIPLICITY :

  procedure Estimate_Multiplicity
               ( p_eval : in Eval_Poly_Sys; j_eval : in Eval_Jaco_Mat;
                 z1,z0,y1,y0 : in Vector; j0,j1 : in Matrix;
                 ipvt1 : in Standard_Natural_Vectors.Vector;
                 orders : in out Vector; fail : out boolean ) is

  -- DESCRIPTION :
  --   [F(z0) - f'(z0)[f'(z1)]^(-1) F(z1)] new_orders = F(z0) old_orders,
  --      with f = polynomial system, f = (f1,f2,..fn), f' its Jacobian,
  --       and F = diag(f1,f2,..,fn),
  --   is the formula for estimating the new orders of multiplicities.

  -- ON ENTRY :
  --   z1        current solution vector;
  --   z0        previous solution vector;
  --   y1        evaluation of z1 in the polynomial system;
  --   y0        evaluation of z0 in the polynomial system;
  --   j0        Jacobian matrix at z0;
  --   j1        lu factorization of the Jacobian matrix at z1;
  --   ipvt1     pivoting information from j1 in the lu factorization;
  --   orders    current estimates for the orders of multiplicities.

  -- ON RETURN :
  --   orders    new estimates for the orders of multiplicities;
  --   fail      true if the matrices become too singular.

  -- ALGORITHM FOR EVALUATION OF THE FORMULA :
  --   A = f'(z0)[f'(z1)]^(-1) F(z1) = f'(z0) B
  --   then B = [f'(z1)]^(-1) F(z1), so we have four steps :
  --   1) solve the matrix equation f'(z1) B = F(z1);
  --   2) compute the matrix product A := f'(z0) B;
  --   3) compute the matrix difference A := F(z0) - A;
  --   4) solve the system A new_orders = F(z0) old_orders.

    n : constant natural := z0'last;
    ipvt : Standard_Natural_Vectors.Vector(1..n);
    info : integer;
    A,B : Matrix(1..n,1..n);
    rhs : Vector(1..n);

  begin
    for i in 1..n loop                    -- solve f'(z1) B = F(z1)
      for j in 1..n loop
        if j /= i
         then rhs(j) := Create(0);
         else Copy(y1(i),rhs(i));
        end if;
      end loop;
      lusolve(j1,n,ipvt1,rhs);
      for j in 1..n loop                  -- update the i-th column of B
        Copy(rhs(j),B(j,i));
      end loop;
      Clear(rhs);
    end loop;
    A := j0*B;                            -- matrix product to compute A
    Clear(B);
    for i in 1..n loop                    -- A := F(z0) - A
      for j in 1..n loop
        Min(A(i,j));
        if i = j
         then Add(A(i,i),y0(i));
        end if;
      end loop;
    end loop;
    lufac(A,n,ipvt,info);                 -- solve A*orders = F(z0)*orders
    if info /= 0
     then fail := true;
     else fail := false;
          for i in 1..n loop
            Mul(orders(i),y0(i));
          end loop;
          lusolve(A,n,ipvt,orders);
    end if;
    Clear(A);
  end Estimate_Multiplicity;

-- TARGET ROUTINES, NEWTON's METHOD :

  procedure Silent_Newton
               ( p_eval : in Eval_Poly_Sys; j_eval : in Eval_Jaco_Mat;
                 zero : in out Solution; epsxa,epsfa : in Floating_Number; 
                 numit : in out natural; max : in natural;
                 fail : out boolean ) is

    n : natural := p_eval'length;
    jacobian : Matrix(1..n,1..n);
    ipvt : Standard_Natural_Vectors.Vector(1..n);
    info : integer;
    y,deltax : Vector(1..n);
    tmp : Floating_Number;

  begin
    y := Eval(p_eval,zero.v);               -- y = f(zero)
    for i in 1..max loop
      jacobian := Eval(j_eval,zero.v);      -- solve jacobian*deltax = -f(zero)
     -- lufco(jacobian,n,ipvt,zero.rco);
      lufac(jacobian,n,ipvt,info);
     -- if Equal(zero.rco,0.0)
      if info /= 0
       then tmp := Sum_Norm(y);
            fail := (tmp > epsfa);
            Clear(tmp);
            return;                         -- singular Jacobian matrix
      end if;
      deltax := -y; Clear(y);
      lusolve(jacobian,n,ipvt,deltax);  
      Add(zero.v,deltax);                   -- make the updates
      y := Eval(p_eval,zero.v);
      Clear(zero.err);              Clear(zero.res);
      zero.err := Sum_Norm(deltax); zero.res := Sum_Norm(y);
      Clear(jacobian); Clear(deltax); 
     -- Clear(y);
      numit := numit + 1;
      if ( zero.err < epsxa ) or else ( zero.res < epsfa ) 
                                            -- stopping criteria
       then fail := false; exit;
       elsif numit >= max
           then fail := true; exit;
      end if;
    end loop;
    jacobian := Eval(j_eval,zero.v);        -- compute condition number
    lufco(jacobian,n,ipvt,zero.rco);
    Clear(jacobian);
  exception
    when numeric_error | constraint_error => fail := true; return;
  end Silent_Newton;

  procedure Silent_Newton
               ( p_eval : in Eval_Poly_Sys; j_eval : in Eval_Jaco_Mat;
                 zero : in out Solution; epsxa,epsfa : in Floating_Number; 
                 orders : in out Vector; estord : in boolean;
                 numit : in out natural; max : in natural;
                 fail : out boolean ) is

    n : natural := p_eval'length;
    jacobian,jaczero,newjac : Matrix(1..n,1..n);
    ipvt : Standard_Natural_Vectors.Vector(1..n);
    y,deltax,prevzero,prevy : Vector(1..n);
    tmp : Floating_Number;

  begin
    y := Eval(p_eval,zero.v);               -- y = f(zero)
    if estord
     then Copy(zero.v,prevzero); Copy(y,prevy);
    end if;
    jacobian := Eval(j_eval,zero.v);        -- solve jacobian*deltax = -f(zero)
    if estord
     then Copy(jacobian,jaczero);
    end if;
    lufco(jacobian,n,ipvt,zero.rco);
    for i in 1..max loop
      if Equal(zero.rco,0.0)
       then tmp := Sum_Norm(y);
            fail := (tmp > epsfa);
            Clear(tmp);
            return;                         -- singular Jacobian matrix
      end if;
      deltax := -y; Clear(y);
      for i in 1..n loop
        Mul(deltax(i),orders(i));
      end loop;
      lusolve(jacobian,n,ipvt,deltax);  
      Add(zero.v,deltax);                   -- make the updates
      y := Eval(p_eval,zero.v);
      Clear(zero.err);              Clear(zero.res);
      zero.err := Sum_Norm(deltax); zero.res := Sum_Norm(y);
      Clear(jacobian); Clear(deltax); 
      numit := numit + 1;
      if ( zero.err < epsxa ) or else ( zero.res < epsfa ) 
                                            -- stopping criteria
       then fail := false; exit;
       elsif numit >= max
           then fail := true; exit;
      end if;
      jacobian := Eval(j_eval,zero.v);
      if estord and then (i < max)
       then Copy(jacobian,newjac);
      end if;
      lufco(jacobian,n,ipvt,zero.rco);
      if estord
       then Estimate_Multiplicity
              (p_eval,j_eval,zero.v,prevzero,y,prevy,jaczero,
               jacobian,ipvt,orders,fail);
            if i < max
             then Copy(zero.v,prevzero); Copy(y,prevy); Copy(newjac,jaczero);
            end if;
      end if;
    end loop;
    if estord
     then Clear(prevzero); Clear(prevy); Clear(jaczero); Clear(newjac);
    end if;
    Clear(y); Clear(jacobian);
  exception
    when numeric_error | constraint_error => fail := true; return;
  end Silent_Newton;

  procedure Reporting_Newton
               ( file : in file_type;
                 p_eval : in Eval_Poly_Sys; j_eval : in Eval_Jaco_Mat;
                 zero : in out Solution; epsxa,epsfa : in Floating_Number; 
                 numit : in out natural; max : in natural;
                 fail : out boolean ) is

    n : natural := p_eval'length;
    jacobian : Matrix(1..n,1..n);
    ipvt : Standard_Natural_Vectors.Vector(1..n);
    info : integer;
    y,deltax : Vector(1..n);
    tmp : Floating_Number;

  begin
    y := Eval(p_eval,zero.v);              -- y = f(zero)
    for i in 1..max loop
      jacobian := Eval(j_eval,zero.v);     -- solve jacobian*deltax = -f(zero)
      lufac(jacobian,n,ipvt,info);
      if info /= 0
       then tmp := Sum_Norm(y);
            fail := (tmp > epsfa);         -- accuracy not reached yet
            Clear(tmp);
            Clear(jacobian); Clear(y);
            return;
      end if;
      deltax := -y; Clear(y);
      lusolve(jacobian,n,ipvt,deltax);  
      Add(zero.v,deltax);                  -- make the updates
      y := Eval(p_eval,zero.v);
      Clear(zero.err);              Clear(zero.res);
      zero.err := Sum_Norm(deltax); zero.res := Sum_Norm(y);
      Clear(deltax); Clear(jacobian);
      numit := numit + 1;
      put(file,"Step "); put(file,numit,4); new_line(file);      -- output
      put(file," |errxa| : "); put(file,zero.err); new_line(file);
      put(file," |errfa| : "); put(file,zero.res); new_line(file);
      if ( zero.err < epsxa ) or else ( zero.res < epsfa ) 
                                                  -- stopping criteria
       then fail := false; exit;
       elsif numit >= max
           then fail := true; exit;
      end if;
    end loop;
    Clear(y);
    jacobian := Eval(j_eval,zero.v);              -- compute condition number
    lufco(jacobian,n,ipvt,zero.rco);
    Clear(jacobian);
  exception
    when numeric_error | constraint_error => fail := true; return;
  end Reporting_Newton;

  procedure Reporting_Newton
               ( file : in file_type;
                 p_eval : in Eval_Poly_Sys; j_eval : in Eval_Jaco_Mat;
                 zero : in out Solution; epsxa,epsfa : in Floating_Number; 
                 orders : in out Vector; estord : in boolean;
                 numit : in out natural; max : in natural;
                 fail : out boolean ) is

    n : natural := p_eval'length;
    jacobian,jaczero,newjac : Matrix(1..n,1..n);
    ipvt : Standard_Natural_Vectors.Vector(1..n);
    y,deltax,prevzero,prevy : Vector(1..n);
    tmp : Floating_Number;

  begin
    y := Eval(p_eval,zero.v);              -- y = f(zero)
    if estord
     then Copy(zero.v,prevzero); Copy(y,prevy);
    end if;
    jacobian := Eval(j_eval,zero.v);       -- solve jacobian*deltax = -f(zero)
    if estord
     then Copy(jacobian,jaczero);
    end if;
    lufco(jacobian,n,ipvt,zero.rco);
    for i in 1..max loop
      if Equal(zero.rco,0.0)               -- singular jacobian matrix
       then tmp := Sum_Norm(y);
            fail := (tmp > epsfa);         -- accuracy not reached yet
            Clear(tmp);
            return;
      end if;
      deltax := -y; --Clear(y);
      for i in 1..n loop
        Mul(deltax(i),orders(i));
      end loop;
      lusolve(jacobian,n,ipvt,deltax);  
      Add(zero.v,deltax);                  -- make the updates
      y := Eval(p_eval,zero.v);
      Clear(zero.err);              Clear(zero.res);
      zero.err := Sum_Norm(deltax); zero.res := Sum_Norm(y);
      Clear(deltax); Clear(jacobian);
      numit := numit + 1;
      put(file,"Step "); put(file,numit,4); new_line(file);      -- output
      put(file," |errxa| : "); put(file,zero.err); new_line(file);
      put(file," |errfa| : "); put(file,zero.res); new_line(file);
      if estord
       then put_line(file," orders : ");
            for i in orders'range loop
              put(file,"   "); put(file,i,1); put(file," : ");
              put(file,orders(i)); new_line(file);
            end loop;
      end if;
      if ( zero.err < epsxa ) or else ( zero.res < epsfa ) 
                                                  -- stopping criteria
       then fail := false; exit;
       elsif numit >= max
           then fail := true; exit;
      end if;
      jacobian := Eval(j_eval,zero.v);
      if estord and then (i < max)
       then Copy(jacobian,newjac);
      end if;
      lufco(jacobian,n,ipvt,zero.rco);
      if estord
       then Estimate_Multiplicity
              (p_eval,j_eval,zero.v,prevzero,y,prevy,jaczero,
               jacobian,ipvt,orders,fail);
            if i < max
             then Copy(zero.v,prevzero); Copy(y,prevy); Copy(newjac,jaczero);
            end if;
      end if;
    end loop;
    Clear(y);
    if estord
     then Clear(prevzero); Clear(prevy); Clear(jaczero); Clear(newjac);
    end if;
    Clear(jacobian);
  exception
    when numeric_error | constraint_error => fail := true; return;
  end Reporting_Newton;

-- APPLICATION OF NEWTON's METHOD TO LISTS OF SOLUTIONS :

  procedure Silent_Root_Refiner
               ( p : in Poly_Sys; sols : in out Solution_List;
                 epsxa,epsfa,tolsing : in Floating_Number;
                 numit : in out natural; max : in natural;
                 estmul : in boolean  ) is

    n : natural := p'length;
    p_eval : Eval_Poly_Sys(1..n) := Create(p);
    jac : Jaco_Mat(1..n,1..n) := Create(p);
    jac_eval : Eval_Jaco_Mat(1..n,1..n) := Create(jac);
    numb : natural;
    fail : boolean;
    sa : Solution_Array(1..Length_Of(sols)) := Create(sols);
    eva : Floating_Number;
    one : Complex_Number;
    eval_acc,orders : Vector(1..n);

  begin
    for i in sa'range loop
      numb := 0;
     -- Clear(sa(i).res);
      eval_acc := Eval(p_eval,sa(i).v);
     -- sa(i).res := Sum_Norm(eval_acc);
      eva := Sum_Norm(eval_acc);
      Clear(eval_acc);
     -- if sa(i).res < 1.0
      if eva < 1.0
       then if estmul
             then one := Create(1);
                  for i in 1..n loop
                    Copy(one,orders(i));
                  end loop;
                  Clear(one);
                  Silent_Newton
                    (p_eval,jac_eval,sa(i).all,epsxa,epsfa,orders,true,
                     numb,max,fail);
             else Silent_Newton
                    (p_eval,jac_eval,sa(i).all,epsxa,epsfa,numb,max,fail);
            end if;
       else fail := true;
      end if;
      Root_Accounting(sa(i),i,sa(sa'first..i),fail,tolsing,epsxa);
      numit := numit + numb;
      Clear(eva);
    end loop;
    Clear(p_eval); Clear(jac); Clear(jac_eval);
    Shallow_Clear(sols); sols := Create(sa); -- Shallow_Clear(sa);
  end Silent_Root_Refiner;

  procedure Silent_Root_Refiner
               ( p : in Poly_Sys; sols,refsols : in out Solution_List;
                 epsxa,epsfa,tolsing : in Floating_Number;
                 numit : in out natural; max : in natural;
                 estmul : in boolean ) is

    n : natural := p'length;
    p_eval : Eval_Poly_Sys(1..n) := Create(p);
    jac : Jaco_Mat(1..n,1..n) := Create(p);
    jac_eval : Eval_Jaco_Mat(1..n,1..n) := Create(jac);
    numb : natural;
    fail : boolean;
    sa : Solution_Array(1..Length_Of(sols)) := Create(sols);
    eva : Floating_Number;
    one : Complex_Number;
    refsols_last : Solution_List;
    eval_acc,orders : Vector(1..n);

  begin
    for i in sa'range loop
      numb := 0;
     -- Clear(sa(i).res);
      eval_acc := Eval(p_eval,sa(i).v);
     -- sa(i).res := Sum_Norm(eval_acc);
      eva := Sum_Norm(eval_acc);
      Clear(eval_acc);
     -- if sa(i).res < 1.0
      if eva < 1.0 then
        if estmul then
          one := Create(1);
          for i in 1..n loop
            Copy(one,orders(i));
          end loop;
          Clear(one);
          Silent_Newton
            (p_eval,jac_eval,sa(i).all,epsxa,epsfa,orders,true,numb,max,fail);
        else
          Silent_Newton
            (p_eval,jac_eval,sa(i).all,epsxa,epsfa,numb,max,fail);
        end if;
      else
        fail := true;
      end if;
      Root_Accounting(sa(i),i,sa(sa'first..i),fail,tolsing,epsxa);
      if not fail
       then Append(refsols,refsols_last,sa(i).all);
      end if;
      numit := numit + numb;
      Clear(eva);
    end loop;
    Clear(p_eval); Clear(jac); Clear(jac_eval);
    Shallow_Clear(sols); sols := Create(sa); -- Shallow_Clear(sa);
  end Silent_Root_Refiner;

  procedure Reporting_Root_Refiner
               ( file : in file_type;
                 p : in Poly_Sys; sols : in out Solution_List;
                 epsxa,epsfa,tolsing : in Floating_Number;
                 numit : in out natural; max : in natural;
                 estmul,wout : in boolean ) is

    n : natural := p'length;
    p_eval : Eval_Poly_Sys(1..n) := Create(p);
    jac : Jaco_Mat(1..n,1..n) := Create(p);
    jac_eval : Eval_Jaco_Mat(1..n,1..n) := Create(jac);
    numb : natural;
    nbfail,nbreg,nbsing,nbclus,nbreal,nbcomp : natural := 0;
    nbtot : constant natural := Length_Of(sols);
    fail : boolean;
    sa : Solution_Array(1..nbtot) := Create(sols);
    initres : Standard_Floating_Vectors.Vector(sa'range);
    eval_acc,orders : Vector(1..n);
    eva : Floating_Number;
    one : Complex_Number;
    t_err,t_rco,t_dis,t_res : Standard_Natural_Vectors.Vector(0..20)
                            := Multprec_Condition_Tables.Create(20); 

  begin
    new_line(file);
    put_line(file,"THE SOLUTIONS :"); new_line(file);
    put(file,nbtot,1); put(file," "); put(file,n,1); new_line(file);
    Write_Bar(file);
    for i in sa'range loop 
      numb := 0;
     -- Clear(sa(i).res);
      eval_acc := Eval(p_eval,sa(i).v);
     -- sa(i).res := Sum_Norm(eval_acc);
      eva := Sum_Norm(eval_acc);
      Clear(eval_acc);
     -- initres(i) := Trunc(sa(i).res);
      initres(i) := Trunc(eva);
     -- if sa(i).res < 1.0
      if eva < 1.0
       then if estmul
             then one := Create(1);
                  for i in 1..n loop
                    Copy(one,orders(i));
                  end loop;
                  Clear(one);
                  if wout
                   then Reporting_Newton
                          (file,p_eval,jac_eval,sa(i).all,epsxa,epsfa,
                           orders,true,numb,max,fail);
                   else Silent_Newton
                          (p_eval,jac_eval,sa(i).all,epsxa,epsfa,
                           orders,true,numb,max,fail);
                  end if;
             elsif wout
                 then Reporting_Newton
                        (file,p_eval,jac_eval,sa(i).all,epsxa,epsfa,
                         numb,max,fail);
                 else Silent_Newton
                        (p_eval,jac_eval,sa(i).all,epsxa,epsfa,numb,max,fail);
            end if;
       else fail := true;
      end if;
      Write_Info(file,sa(i).all,initres(i),i,numb,fail);
      Root_Accounting(file,sa(i),i,sa(sa'first..i),fail,tolsing,epsxa,
                      nbfail,nbreal,nbcomp,nbreg,nbsing,nbclus);
      Update_Corrector(t_err,sa(i).all);
      Update_Condition(t_rco,sa(i).all);
      Update_Residuals(t_res,sa(i).all);
      numit := numit + numb;
      Clear(eva);
    end loop;
    Write_Global_Info(file,nbtot,nbfail,nbreal,nbcomp,nbreg,nbsing,nbclus);
    Clear(p_eval); Clear(jac); Clear(jac_eval);
    Shallow_Clear(sols); sols := Create(sa); -- Shallow_Clear(sa);
    Update_Distances(t_dis,sols);
    Write_Tables(file,t_err,t_res,t_rco,t_dis);
  end Reporting_Root_Refiner;

  procedure Reporting_Root_Refiner
               ( file : in file_type;
                 p : in Poly_Sys; sols,refsols : in out Solution_List;
                 epsxa,epsfa,tolsing : in Floating_Number;
                 numit : in out natural; max : in natural;
                 estmul,wout : in boolean ) is

    n : natural := p'length;
    p_eval : Eval_Poly_Sys(1..n) := Create(p);
    jac : Jaco_Mat(1..n,1..n) := Create(p);
    jac_eval : Eval_Jaco_Mat(1..n,1..n) := Create(jac);
    numb : natural;
    nbfail,nbreg,nbsing,nbclus,nbreal,nbcomp : natural := 0;
    nbtot : constant natural := Length_Of(sols);
    fail : boolean;
    sa : Solution_Array(1..nbtot) := Create(sols);
    initres : Standard_Floating_Vectors.Vector(sa'range);
    refsols_last : Solution_List;
    eval_acc,orders : Vector(1..n);
    eva : Floating_Number;
    one : Complex_Number;
    t_err,t_rco,t_dis,t_res : Standard_Natural_Vectors.Vector(0..20)
                            := Multprec_Condition_Tables.Create(20); 

  begin
    new_line(file);
    put_line(file,"THE SOLUTIONS :"); new_line(file);
    put(file,nbtot,1); put(file," "); put(file,n,1); new_line(file);
    Write_Bar(file);
    for i in sa'range loop 
      numb := 0;
     -- Clear(sa(i).res);
      eval_acc := Eval(p_eval,sa(i).v);
     -- sa(i).res := Sum_Norm(eval_acc);
      eva := Sum_Norm(eval_acc);
      Clear(eval_acc);
     -- initres(i) := Trunc(sa(i).res);
      initres(i) := Trunc(eva);
     -- if sa(i).res < 1.0
      if eva < 1.0 then
        if estmul then
          one := Create(1);
          for i in 1..n loop
            Copy(one,orders(i));
          end loop;
          Clear(one);
          if wout then
            Reporting_Newton(file,p_eval,jac_eval,sa(i).all,epsxa,epsfa,
                             orders,true,numb,max,fail);
          else
            Silent_Newton(p_eval,jac_eval,sa(i).all,epsxa,epsfa,
                          orders,true,numb,max,fail);
          end if;
        elsif wout then
          Reporting_Newton(file,p_eval,jac_eval,sa(i).all,epsxa,epsfa,
                           numb,max,fail);
        else
          Silent_Newton(p_eval,jac_eval,sa(i).all,epsxa,epsfa,numb,max,fail);
        end if;
      else
        fail := true;
      end if;
      Write_Info(file,sa(i).all,initres(i),i,numb,fail);
      Root_Accounting(file,sa(i),i,sa(sa'first..i),fail,tolsing,epsxa,
                      nbfail,nbreal,nbcomp,nbreg,nbsing,nbclus);
      if not fail
       then Append(refsols,refsols_last,sa(i).all);
      end if;
      Update_Corrector(t_err,sa(i).all);
      Update_Condition(t_rco,sa(i).all);
      Update_Residuals(t_res,sa(i).all);
      numit := numit + numb;
      Clear(eva);
    end loop;
    Write_Global_Info(file,nbtot,nbfail,nbreal,nbcomp,nbreg,nbsing,nbclus);
    Clear(p_eval); Clear(jac); Clear(jac_eval);
    Shallow_Clear(sols); sols := Create(sa); -- Shallow_Clear(sa);
    Update_Distances(t_dis,sols);
    Write_Tables(file,t_err,t_res,t_rco,t_dis);
  end Reporting_Root_Refiner;

end Multprec_Root_Refiners;
