with Standard_Natural_Vectors;          use Standard_Natural_Vectors;

package Checker_Moves is

-- DESCRIPTION :
--   This package defines the primitive operations to move
--   black and white checkers in a combinatorial game to count
--   the Littlewood-Richardson coefficients.

-- PART I : Moves of the Black Checkers 

  function Identity_Permutation ( n : natural ) return Vector;

  -- DESCRIPTION :
  --   Returns the identity permutation of n elements.

  function Descending_Checker ( p : Vector ) return natural;

  -- DESCRIPTION :
  --   Returns either the index of the descending black checker,
  --   or zero if the permutation is already the longest word.

  function Rising_Checker ( p : Vector; i : natural ) return natural;

  -- DESCRIPTION :
  --   Given the index i to the descending black checker in p,
  --   returns the index to the rising checker.
  --   If i = 0, then 0 is returned.

  procedure Move ( p : in out Vector; down,up : in natural );

  -- DESCRIPTION :
  --   Performs the move on the permutation p, with the index 
  --   of the descending checker in down, and the index of the
  --   rising checker in up.

  function Number_of_Moves ( n : natural ) return natural;

  -- DESCRIPTION :
  --   Returns the number of moves made by n black checkers.

-- PART II : Moves of the White Checkers

  function Critical_Row ( r,c : natural; rows,cols : Vector ) return natural;

  -- DESCRIPTION :
  --   Returns in the index to entry of rows of the white checker
  --   in the critical row r and column c.  If there are no white 
  --   checkers in the critical row, then the value on return is zero.

  function In_Critical_Diagonal 
              ( dr,dc,n,row,column : natural ) return boolean;

  -- DESCRIPTION :
  --   Returns true if there is a k, dr + k <= n and dc + k <= n, 
  --   such that row = dr + k and column = dc + k;
  --   returns false otherwise.

  -- ON ENTRY :
  --   (dr,dc)  start of the critical diagonal;
  --   n        size of the checkers board;
  --   row      row of a checker;
  --   column   column of a checker.

  function Top_White_Checker
              ( dr,dc,n : natural; rows,cols : Vector ) return natural;

  -- DESCRIPTION :
  --   Returns the index of the highest checker in the critical diagonal.
  --   Returns 0 if there is no checker in the critical diagonal.

  function Between ( a,b,x : natural ) return boolean;

  -- DESCRIPTION :
  --   Returns true if x lies between a and b.
  --   Notice that a and b are not necessarily sorted.

  function Blocker ( rows,cols : Vector; t,r : natural ) return boolean;

  -- DESCRIPTION :
  --   Returns true if there is a blocker between the top white checker
  --   indexed by t and the checker indexed by the row r.

  procedure Swap ( rows : in out Vector; i,j : in natural );

  -- DESCRIPTION :
  --   If i /= 0, then the rows(i) is swapped with rows(j).

  function Central_Choice
              ( p : Vector; d,r : natural;
                rows,cols : Vector; cr,cd : natural; debug : boolean ) 
              return natural;

  -- DESCRIPTION :
  --   Determines whether to stay, swap, or whether to do both,
  --   depending on the position of the white checker in the
  --   critical row and the top white checker in the critical diagonal.

  -- ON ENTRY :
  --   p        permutation determines black checkers;
  --   d        descending black checker is at (p(d),p'last-d+1);
  --   r        rising black checker is at (p(r),p'last-r+1);
  --   rows     rows of the white checkers;
  --   cols     cols(cols'last-i+1) is column for checker in rows(i);
  --   cr       index to rows for white checker in critical row;
  --   cd       index to cols for top white checker in critical diagonal;
  --   debug    if true, then additional information is printed;
  --            otherwise, there is no additional output.

  -- REQUIRED :
  --   d,r,cr,cd are valid entries in the arrays (all > 0).

  -- ON RETURN :
  --   0        if the white checkers must stay;
  --   1        if the white checkers must swap;
  --   2        if choice between stay or swap.

  function Happy_in_Row ( p : Vector; r,c : natural ) return boolean;

  -- DESCRIPTION :
  --   Returns true if there is a black checker to the left of
  --   the white checker in row r and column c.

  function Happy_in_Column ( p : Vector; r,c : natural ) return boolean;

  -- DESCRIPTION :
  --   Returns true if there is a black checker above the
  --   white checker in row r and column c.

  procedure Make_Happy ( p : in Vector; rows,cols : in out Vector );

  -- DESCRIPTION :
  --   Shifts white checkers to the left or up until they are happy,
  --   or cannot be shifted any further.

end Checker_Moves;
