/* The following is a simple interface of MPI with PHCpack:
 * after prompting the user for a target and start system,
 * the homotopy is then broadcasted to all the nodes
 * dynamically. */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "mpi.h"

#define SEND_SSOL 100 /* tag for sending start solution */
#define SEND_SMUL 101 /* tag for sending multiplicity of start solution */
#define SEND_TSOL 102 /* tag for sending target solution */
#define SEND_TMUL 103 /* tag for sending multiplicity of target solution */

#define v 0  /* verbose flag */

extern void adainit( void );
extern int _ada_use_c2phc ( int task, int *a, int *b, double *c );
extern void adafinal( void );

void dimension_broadcast ( int myid, int *n );
/* gets the dimension n from the root node,
 * broadcasts the dimension then to all the nodes, and
 * initializes the systems container at each node */

void monomials_broadcast ( int myid, int n );
/* broadcasts all monomials in the system to the nodes,
 * adding the monomials to the containers at each node */

void path_track(int m[2], double* sol);
/* do path tracking for the given start solution */

void dynamic_load(int myid, int n, int numprocs, int nbsols, int *mysolnb);
/* the root node distributes one start solution to other nodes each
 * time. after finshing path tracking, the nodes return their target
 * solutions to the root node. the root node prints out the solution 
 * and distributes a new job to the node which finishes its job    */ 

void print_solution(int n, int m, double *sol);
/* the root node prints out the target solutions */

void copy_broadcast ( int myid );
/* copies the result of the broadcast to PHCpack data and
 * prints the result of the broadcast at the node */

void start_system_broadcast ( int myid, int n, int *nbsols );
/* broadcast of start system and start solutions to nodes,
 * n is the dimension and nbsols the number of solutions */

void print_monomials ( void );
/* prints the monomials in the container */

void print_system ( void );
/* prints the polynomial system in the container */

void print_homotopy ( void );
/* prints the target and start system in the homotopy */

void print_time(double *time, int numprocs, int *mysolnb);
/* prints the time spent and the number of paths tracked on 
 * each of the processors */

int main ( int argc, char *argv[] )
{
   int myid,numprocs,n,nbsols,mysolnb,*mysol;
   double startwtime,endwtime,wtime,*time;
   MPI_Status status;
 
   adainit();
   MPI_Init(&argc,&argv);
   MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
   MPI_Comm_rank(MPI_COMM_WORLD,&myid);

   if(myid ==0)
   {
      time = (double*) calloc(numprocs, sizeof(double));
      mysol = (int*) calloc(numprocs, sizeof(int));
      startwtime = MPI_Wtime();
   }

   dimension_broadcast(myid,&n);
   monomials_broadcast(myid,n);
   copy_broadcast(myid); 
   start_system_broadcast(myid,n,&nbsols);

   MPI_Bcast(&nbsols,1,MPI_INT,0,MPI_COMM_WORLD);
   
   if(myid!=0)
      startwtime = MPI_Wtime();
   dynamic_load(myid, n, numprocs, nbsols, &mysolnb);
   if(myid!=0)
      endwtime = MPI_Wtime();

   MPI_Barrier(MPI_COMM_WORLD);
   if(myid == 0) endwtime = MPI_Wtime();
   wtime = endwtime-startwtime;
   MPI_Gather(&wtime,1,MPI_DOUBLE,time,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Gather(&mysolnb,1,MPI_INT,mysol,1,MPI_INT,0,MPI_COMM_WORLD);
   if(myid == 0) print_time(time,numprocs,mysol);
   MPI_Finalize();
   adafinal();

   return 0;
}

void dimension_broadcast ( int myid, int *n )
{
   int *d,fail,m;
   double *c;   

   if(myid == 0)
   {
      fail = _ada_use_c2phc(11,n,d,c);    /* read target system */
     /* fail = _ada_use_c2phc(12,n,d,c); */   /* write target system */
      fail = _ada_use_c2phc(1,n,d,c);     /* copy target to container */
     /* fail = _ada_use_c2phc(21,n,d,c); */   /* write system in container */
      fail = _ada_use_c2phc(22,n,d,c);    /* get dimension */
     /* printf("The dimension is %d.\n", *n); */
   }
 
   MPI_Bcast(n, 1, MPI_INT, 0, MPI_COMM_WORLD);  

   if(myid != 0)
   {
      /* printf("Node %d knowns that the dimension is %d\n", myid, *n); */
      fail = _ada_use_c2phc(23,n,d,c);    /* initialize container */
      /* fail = _ada_use_c2phc(22,&m,d,c); */ /* get dimension as test */
      /* printf("  and initialized container with dimension %d.\n", m); */
   }
}

void monomials_broadcast ( int myid, int n )
{
   double cff[2];
   int i,j,exp[n],m[3],mm,fail;

   for(i=1; i<=n; i++)
   {
      m[1] = i;
      if(myid == 0)
      {
         fail = _ada_use_c2phc(24,m,exp,cff);    /* get #monomials */
         mm = m[0];
        /* printf("Polynomial %d has %d monomials.\n",i,mm); */
      }

      MPI_Bcast(&mm, 1, MPI_INT, 0, MPI_COMM_WORLD);

      m[0] = n;
      for(j=1; j<=mm; j++)    /* broadcast j-th term of i-th polynomial */
      {
         m[2] = j;
         if(myid == 0) fail = _ada_use_c2phc(25,m,exp,cff); 
         MPI_Bcast(cff,2,MPI_DOUBLE,0,MPI_COMM_WORLD); 
         MPI_Bcast(exp,n,MPI_INT,0,MPI_COMM_WORLD); 
         if(myid != 0) fail = _ada_use_c2phc(26,m,exp,cff); 
      }
   }
}

void dynamic_load(int myid, int n, int numprocs, int nbsols, 
                  int *mysolnb)
{
  int dest, m[2], send[2], i, k, fail;
  double sol[2*n+5];
  MPI_Status status;

  m[0] = n;
  *mysolnb = 0;
  printf("\n");
  if(myid==0)
  {
    /*distributes one start root for each client at the beginning */
    for(dest=1; dest<=numprocs-1; dest++)
    {
       k = dest;
       fail = _ada_use_c2phc(34,&k,&m[1],sol); 
       /* return a solution from the container */
       send[0] = dest; send[1] = m[1];
       MPI_Send(send,2,MPI_INT,dest,SEND_SMUL,MPI_COMM_WORLD);
       MPI_Send(sol,2*n+5,MPI_DOUBLE,dest,SEND_SSOL,MPI_COMM_WORLD);  
       (*mysolnb)++;
    }
    printf("THE SOLUTIONS :\n\n");
    printf("%d %d\n", nbsols, n);
    printf("===========================================================\n");

    /*collets the target roots and distributes the remain start roots */ 
    for(k=numprocs; k<=nbsols+numprocs-1; k++)
    {
       MPI_Recv
        (&m[1],1,MPI_INT,MPI_ANY_SOURCE,SEND_TMUL,MPI_COMM_WORLD,&status);
       MPI_Recv
        (sol,2*n+5,MPI_DOUBLE,MPI_ANY_SOURCE,SEND_TSOL,MPI_COMM_WORLD,&status);

       printf("solution %d :\n", k-numprocs+1);
       print_solution(n, m[1], sol);       /* write the target solutions */

       fail = _ada_use_c2phc(34,&k,&m[1],sol); 
       /* return a solution from the container */
       send[0] = k; send[1] = m[1];
       dest = status.MPI_SOURCE;
       MPI_Send(send,2,MPI_INT,dest,SEND_SMUL,MPI_COMM_WORLD);
       if(k<=nbsols)
       {
         MPI_Send(sol,2*n+5,MPI_DOUBLE,dest,SEND_SSOL,MPI_COMM_WORLD);
         (*mysolnb)++;
       }
    }
  }
  else
  {
    while(1)
    {
       MPI_Recv(send,2,MPI_INT,0,SEND_SMUL,MPI_COMM_WORLD,&status);
       if (send[0]>nbsols) break;
       m[1] = send[1];
       MPI_Recv(sol,2*n+5,MPI_DOUBLE,0,SEND_SSOL,MPI_COMM_WORLD,&status); 
       path_track(m, sol);
       MPI_Send(&m[1],1,MPI_INT,0,SEND_TMUL,MPI_COMM_WORLD);
       MPI_Send(sol,2*n+5,MPI_DOUBLE,0,SEND_TSOL,MPI_COMM_WORLD);
       (*mysolnb)++;
   }
   if(v) printf("Node %d done.\n", myid);
  }
}

void print_solution(int n, int m, double *sol)
{
   int i; 

   printf("t : %.15e  %.15e\n", sol[0], sol[1]);
   printf("m : %d\n", m);
   printf("The solution for t :\n");
   for(i=0; i<n; i++)
      printf(" %.15e  %.15e\n", sol[2*i+2], sol[2*i+3]);
   printf("== err : %.3e = rco : %.3e = res : %.3e ==\n",
           sol[2*n+2],sol[2*n+3],sol[2*n+4]);
}

void path_track(int m[2], double* sol)
{
   int *a,*b,fail,k=1;
   double *c;
 
   fail = _ada_use_c2phc(36,&k,m,sol); /* append a solution to the container */ 
   fail = _ada_use_c2phc(8,a,b,c);     /* copy container to start sols */
   fail = _ada_use_c2phc(37,a,b,c);    /* clear solution container */
   fail = _ada_use_c2phc(16,a,b,c);    /* do path tracking */
   fail = _ada_use_c2phc(5,a,b,c);     /* copy target sols to container */
   fail = _ada_use_c2phc(34,&k,&m[1],sol); 
   /* return a solution from the container */
   fail = _ada_use_c2phc(37,a,b,c);    /* clear solution container */
}

void copy_broadcast ( int myid )
{
   int n,*a,*b,fail;
   double *c;

/*   if(myid == 1)
   {
      printf("The monomials at node %d:\n", myid);
      print_monomials();
      printf("The polynomial system at node %d:\n", myid);
      print_system();
   }
*/
   if(myid != 0)
   {
      fail = _ada_use_c2phc(2,a,b,c);     /* copy container to target */
      fail = _ada_use_c2phc(27,a,b,c);    /* clear systems container */
     /* fail = _ada_use_c2phc(12,a,b,c); */  /* write target system */
   }
}

void start_system_broadcast ( int myid, int n, int *nbsols )
{
   int *a,*b,fail;
   double *c;

   if(myid == 0)
   {
      fail = _ada_use_c2phc(13,a,b,c);    /* read start system */
     /* fail = _ada_use_c2phc(14,a,b,c); */  /* write start system */
      fail = _ada_use_c2phc(3,a,b,c);     /* copy start to container */
     /* fail = _ada_use_c2phc(21,a,b,c); */  /* write system in container */
      fail = _ada_use_c2phc(7,a,b,c);     /* copy start sols to container */
     /* fail = _ada_use_c2phc(15,a,b,c); */  /* write start solutions */
   }
   if(myid != 0)
      fail = _ada_use_c2phc(23,&n,b,c);   /* initialize container */

   monomials_broadcast(myid,n);           /* broadcast container */

   if(myid != 0)                          /* copy result of broadcast */
   {
      fail = _ada_use_c2phc(4,a,b,c);     /* copy container to target */
      fail = _ada_use_c2phc(27,a,b,c);    /* clear systems container */
     /* fail = _ada_use_c2phc(14,a,b,c); */  /* write start system */
   }

   if(myid == 0)
   {
      fail = _ada_use_c2phc(32,a,nbsols,c);  /* get #solutions */
     /* printf("There are %d start solutions.\n", *nbsols); */
   }
}

void print_monomials ( void ) 
{
   int *d,i,j,k,n,m[3],mm,fail;
   double c[2];

   fail = _ada_use_c2phc(22,&n,d,c);

   d = (int*)calloc(n,sizeof(int));

   for(i=1; i<=n; i++)
   {
      m[1] = i;
      fail = _ada_use_c2phc(24,m,d,c);
      mm = m[0];
      m[0] = n;
      printf("Polynomial %d has %d monomials :\n",i,mm);
      for(j=1; j<=mm; j++)
      {
         m[2] = j;
         fail = _ada_use_c2phc(25,m,d,c);
         printf(" %.15e  %.15e",c[0],c[1]);
         for(k=0; k<n; k++) printf(" %d",d[k]);
         printf("\n");
      }
   }
}

void print_system ( void ) 
{
   int *a,*b,fail;
   double *c;

   fail = _ada_use_c2phc(21,a,b,c);
}

void print_homotopy ( void )
{
   int *a,*b,fail;
   double *c;

   printf("\n");
   fail = _ada_use_c2phc(12,a,b,c);
   fail = _ada_use_c2phc(14,a,b,c);
   printf("The start solutions :\n");
   fail = _ada_use_c2phc(15,a,b,c);
}

void print_time(double *time, int numprocs, int *mysolnb)
{
   int i;
   printf("\nTotal wall time = %lf seconds on %d processors\n",
          time[0],numprocs);
   printf("Total number of paths tracked: %d.\n", mysolnb[0]);

   for(i=1; i<numprocs; i++)
   {
      printf("The wall time spent on NO. %d processor = %lf seconds.\n",
             i,time[i]);
      printf("The number of paths tracked on NO. %d processor: %d.\n",
             i,mysolnb[i]);
   }
   free(time);
   free(mysolnb);
}
